# -*- coding: utf-8 -*-
"""
author: Bowei Pu at 2023.03.01
version: 2024.01.15
function: Some useful tools
"""
import spglib
from pybw.tricks.lazy_import import *

__author__ = 'Bowei Pu'
__version__ = '2024.01.15'
__maintainer__ = 'Bowei Pu'
__email__ = 'pubowei@foxmail.com'
__status__ = 'Beta'
__date__ = '2024-01-15'


def get_oxi_string(oxi_state):
    """
    Generate the text string of oxidation state information of elements 
    to write in a cif file.
    
    Parameters
    ----------
    oxi_state: dict
        Symbol and oxi_state of elements.
    
    Returns
    -------
    string: str
        The text string of file contents.
    """
    s = '\nloop_\n_atom_type_symbol\n_atom_type_oxidation_number\n'
    for k, v in oxi_state.items():
        v = int(v) if isinstance(v, int) or v.is_integer() else float(v)
        s += '{} {}\n'.format(k, v)
    return s + '\n'


def read_possible_species(species):
    """
    Convert possible_species list string of doc to dict
    
    Args:
        species: possible_species, list of string
                 Eg., ['Li+', 'La3+', 'Zr4+', 'O2-']
    
    Note:
        In some possible_species, one element may have more than one possible
        specie, this will make problem when apply it to add stru's oxi state. 
        Thus, the valence given by possible_species should be used with caution.
        Eg., mp-761342 'Li4 Co2 Cu3 Sb3 O16', possible_species is ['Li+', 
        'Sb5+', 'O2-', 'Cu2+', 'Cu3+', 'Co3+']. Here, Cu has two oxi state +2
        and +3
    """
    oxi = {}
    for specie in species:
        elem = re.findall('[a-zA-Z]+', specie)[0]
        sign = re.findall('[+-]', specie)[0]
        numb = re.findall('\d+', specie)
        numb = numb[0] if numb else '1'
        valence = Decimal(sign + numb)
        oxi[elem] = valence
    return oxi


class MPR():
    """
    Easy usage of MPRester
    """
    def __init__(self, key=''):
        self.key = (self.initial_mp_api_key() if self.initial_mp_api_key() 
                    else '2Nix6DqUWNr8vXUGlh93rKXlTwUbF0IR')
        
        self.mpr = MPRester(self.key)
        self.database_version = self.mpr.get_database_version()
    
    def search(self, **kwargs):
        try:
            docs = self.mpr.materials.summary.search(**kwargs)
        except:
            docs = self.mpr.summary.search(**kwargs)
        docs.sort(key=lambda x: int(x.material_id.split('-')[1]))
        docs = [DictDoc(i.dict()) for i in docs]
        for doc in docs:
            doc.setattr('database_version', self.database_version)
        return docs

    @classmethod
    def initial_mp_api_key(csl, key: str=''):
        key = str(key).strip()
        if not key:
            key_files = []
            for p in ['C:/', os.path.expanduser('~'), 
                    os.path.expanduser('~') + '/Desktop']:
                for file in os.listdir(p):
                    file = os.path.join(p, file)
                    if os.path.isfile(file):
                        if re.search('mp_api_key', file.lower()):
                            key_files.append(file)
            key_files.sort(key=lambda x: os.path.getmtime(x), reverse=True)
            keys = []
            for file in key_files:
                with open(file, 'r') as f:
                    c = [i.strip() for i in f.readlines() if len(i.strip()) == 32]
                    keys.extend(c)
            if not keys:
                key = '2Nix6DqUWNr8vXUGlh93rKXlTwUbF0IR'
                return key
            key = keys[0]
            return key
        return key


class DictDoc():
    """
    Dict to class
    Designed to compatible with MPDoc
    """
    def __init__(self, dic: dict={}):
        self.dic = dic
        self._generate_attributes()
    
    def __repr__(self):
        if hasattr(self, 'structure'):
            return '<DictDoc: {}>'.format(self.structure.formula)
        else:
            return '<DictDoc: >'
    
    def __str__(self):
        return self.__repr__()
        
    def dict(self):
        return self.dic
    
    def as_dict(self):
        return self.dic
    
    def keys(self):
        return self.dic.keys()
    
    def values(self):
        return self.dic.values()
    
    def _generate_attributes(self):
        for k, v in self.dic.items():
            setattr(self, k, v)
    
    def setattr(self, name, value):
        self.dic[name] = value
        setattr(self, name, value)
    
    def add(self, name, value):
        self.setattr(name, value)
    
    def get_light_doc(self, reserve=[]):
        reserves = ['formula_pretty', 'formula_anonymous', 'chemsys',
                    'property_name', 'material_id', 'es_source_calc_id',
                    'ordering', 'database_version']
        reserves += ['nsites', 'nelements', 'num_magnetic_sites', 
                     'num_unique_magnetic_sites']
        reserves += ['volume', 'density', 'density_atomic',
                     'uncorrected_energy_per_atom', 'energy_per_atom',
                     'formation_energy_per_atom', 'energy_above_hull',
                     'band_gap', 'cbm', 'vbm', 'efermi',
                     'total_magnetization',
                     'total_magnetization_normalized_vol',
                     'total_magnetization_normalized_formula_units']
        reserves += ['deprecated', 'is_stable', 'is_gap_direct',
                     'is_metal',  'is_magnetic', 'theoretical']
        reserves += ['elements', 'composition', 'composition_reduced', 
                     'decomposes_to', 'symmetry', 'structure', 'task_ids', 
                     'possible_species', 
                     'guess_oxi_states', 'oxi_states']
        reserves += reserve
        return self.__class__({i: self.__getattribute__(i) 
                               for i in reserves if i in self.__dir__()})


class StructureSaveCif():
    """
    Find the symmetry of a structure, and save it's cif file with symmetry.
    
    Attributes
    ----------
    structure: pymatgen Structure
        structure
    symprec: float, int or None
        If not symprec, spg info will be given by guess
    """
    def __init__(self, structure, symprec=None):
        self.structure = structure
        self.symprec, self.spg_no, self.spg_symbol = self._get_symprec_info(symprec)

    def _get_symprec_info(self, symprec):
        if isinstance(symprec, int) or isinstance(symprec, float):
            spg_info = self.structure.get_space_group_info(symprec)
            symprec_info = [symprec, spg_info[1], spg_info[0]]
        else:
            symprec_info = self.get_space_group_info()[0]
        return symprec_info

    def get_space_group_info(self):
        lattice = self.structure.lattice.matrix
        positions = self.structure.frac_coords
        numbers = [site.specie.number for site in self.structure.sites]
        cell = (lattice, positions, numbers)
        symprecs = [0.01] + [round(i, 2) for i in np.arange(0.1, 1.51, 0.1)]
        _spg_info = []
        for i in symprecs:
            _info = spglib.get_symmetry_dataset(cell=cell, symprec=i)
            if not _info:
                break
            _info = [i, _info['number'], _info['international']]
            _spg_info.append(_info)
        spg_info = []
        _spg_num = []
        for i in _spg_info:
            if i[1] not in _spg_num:
                _spg_num.append(i[1])
                spg_info.append(i)
        if spg_info[0][1] == 1 and len(spg_info >= 2):
            spg_info = spg_info[1:]
        return spg_info
    
    def save_cif(self, filename=''):
        if not filename:
            formula = self.structure.formula.replace(' ', '')
            filename = '{}-spg_{}-{}.cif'.format(formula, self.spg_no, 
                                                 time.strftime('%y%m%d_%H%M'))
        self.structure.to(fmt='cif', filename=filename, symprec=self.symprec)






        
        
        
        
        bvs.Rmin_moveion = Rmin_moveion
        
        if len(bvs._Struct._OxidationTable) == 0:
            raise Exception('can not calculate BV and Ea without atom oxi info')
        
        bvs.stop = False
        
        centrepos = np.zeros(bvs._Size + [3])
        bvs.__Data['BVS'] = np.zeros(bvs._Size, dtype=np.double)
        bvs.__Data['BVSE'] = np.zeros(bvs._Size, dtype=np.double)
        bvs.__Data['BVEL'] = np.zeros(bvs._Size, dtype=np.double)
        
        for k in range(bvs._Size[2]):
            for j in range(bvs._Size[1]):
                for i in range(bvs._Size[0]):
                    centrepos[i][j][k][2] = k / (bvs._Size[2] - 1.0)
                    centrepos[i][j][k][1] = j / (bvs._Size[1] - 1.0)
                    centrepos[i][j][k][0] = i / (bvs._Size[0] - 1.0)
        
        if ProgressCall:
            ProgressCall(0)
        
        bvs._poss = bvs._Struct.FracPosToCartPos(centrepos)

        atomsq = {}
        for atom in bvs._Struct._atomsymbols:
            atomsq[atom] = 0
            for site in bvs._Struct._Sites:
                if atom in site.GetElements():
                    key = atom + str(site.GetElementsOxiValue()[atom])
                    atomsq[atom] = atomsq[atom] + site.GetElementsOxiValue()[atom] * site.GetElementsOccupy()[atom] / math.sqrt(bvs._Elementsparam[key][3])
        
        qsumanion = 0
        qsumcation = 0
        for (atom, value) in atomsq.items():
            if value > 0:
                qsumcation = qsumcation + value
            elif value < 0:
                qsumanion = qsumanion + value
            else:
                qsumanion = 0
                qsumcation = 0
        
        qsum = 0.0
        if bvs.ValenceOfMoveIon > 0:
            qsum = -qsumanion / qsumcation
        else:
            qsum = -qsumcation / qsumanion
        
        key1 = bvs._MoveIon + str(bvs.ValenceOfMoveIon)
        qm1 = bvs.ValenceOfMoveIon / math.sqrt(bvs._Elementsparam[key1][3])
        rm1 = bvs._Elementsparam[key1][6]
        
        for i in tqdm(range(bvs._Size[0])):
            (distance, neighborsindex) = bvs._Struct.GetKNeighbors(bvs._poss[i], kn=128)
            
            if ProgressCall:
                ProgressCall(10 + i * 90 / (bvs._Size[0] - 1))
            
            for j in range(bvs._Size[1]):
                
                for k in range(bvs._Size[2]):
                    if bvs.stop:
                        return False
                    
                    bvsdata = 0.0
                    data = 0.0
                    cdata = 0.0
                    bvelcdata = 0.0
                    N = 0
                    D0 = 0.0
                    Rcutoff = 10.0
                    alpha = 0.0
                    occupyvalue = 0.0
                    
                    for dindex, index in enumerate(neighborsindex[j][k]):
                        site2 = bvs._Struct._SuperCellusites[index]
                        
                        if site2.GetIronType() * bvs.ValenceOfMoveIon < 0:
                            for (ssymbol, occupyvalue) in site2.GetElementsOccupy().items():
                                if ssymbol != 'Vac':
                                    site2oxi = site2.GetElementsOxiValue()[ssymbol]
                                    if bvs.ValenceOfMoveIon > 0:
                                        key = "".join([bvs._MoveIon, str(bvs.ValenceOfMoveIon), ssymbol, str(site2oxi)])
                                    else:
                                        key = "".join([ssymbol, str(site2oxi), bvs._MoveIon, str(bvs.ValenceOfMoveIon)])
                                    if (key in bvs._BVSEparam) and (key in bvs._BVparam):
                                        (Nc, r0, Rcutoff, D0, Rmin, alpha) = bvs._BVSEparam[key][0]
                                        (r, b) = bvs._BVparam[key][0]
                                        Rcutoff = 10.0
                                        if distance[j][k][dindex] <= Rcutoff:
                                            # bv=np.exp((r0-distance[k][j][i][dindex])*alpha)
                                            _bvs = np.exp((r - distance[j][k][dindex]) / b)
                                            smin = np.exp((Rmin - distance[j][k][dindex]) * alpha)
                                            en_s = np.exp((Rmin - Rcutoff) * alpha)
                                            bvsdata = bvsdata + _bvs  #*occupyvalue-((en_s-1)**2-1)
                                            data = data + ((smin - 1)**2 - 1) * occupyvalue
                                    else:
                                        if (key not in bvs._BVSEparam):
                                            raise Exception("bvse {0}  param can't find!!!!".format(key))
                                            ProgressCall(0)
                                        else:
                                            raise Exception("_bvs {0} param can't find!!!!".format(key))
                                            ProgressCall(0)
                        else:
                            for (ssymbol, occupyvalue) in site2.GetElementsOccupy().items():
                                if ssymbol != 'Vac':
                                    site2oxi = site2.GetElementsOxiValue()[ssymbol]
                                    
                                    if ssymbol != bvs._MoveIon:
                                        key = ssymbol + str(site2oxi)
                                        rm2 = bvs._Elementsparam[key][6]
                                        qm2 = site2oxi / math.sqrt(bvs._Elementsparam[key][3])
                                        rm1m2 = distance[j][k][dindex]
                                        f = 0.74
                                        if rm1m2 > rm2:
                                            if rm1m2 < Rcutoff:
                                                cdata = cdata + occupyvalue * qm1 * qm2 / rm1m2 * erfc(rm1m2 / (f * (rm1 + rm2))) * qsum
                                                bvelcdata = bvelcdata + occupyvalue * qm1 * qm2 / rm1m2 * (erfc(rm1m2 / (f * (rm1 + rm2))) - erfc(Rcutoff / (f * (rm1 + rm2))))
                                        else:
                                            cdata = 300
                                            bvelcdata = 300
                                    
                                    elif ssymbol == bvs._MoveIon:
                                        key = ssymbol + str(site2oxi)
                                        rm2 = bvs._Elementsparam[key][6]
                                        qm2 = site2oxi / math.sqrt(bvs._Elementsparam[key][3])
                                        rm1m2 = distance[j][k][dindex]
                                        f = 0.74
                                        
                                        if rm1m2 > bvs.Rmin_moveion:
                                            if rm1m2 < Rcutoff:
                                                cdata = cdata + occupyvalue * qm1 * qm2 / rm1m2 * erfc(rm1m2 / (f * (rm1 + rm2))) * qsum
                                                bvelcdata = bvelcdata + occupyvalue * qm1 * qm2 / rm1m2 * (erfc(rm1m2 / (f * (rm1 + rm2))) - erfc(Rcutoff / (f * (rm1 + rm2))))
                                        else:
                                            cdata = cdata + 0
                                            bvelcdata = bvelcdata + 0


# -*- coding: utf-8 -*-
"""
author: Bowei Pu at 2023.07.27
version: 2023.08.09

Convert image files to pdf
"""

import os
import re
from glob import glob
from tqdm import tqdm
from pathlib import Path
from PIL import Image
from joblib import Parallel, delayed


def images_to_pdf(images, pdf_file, use_tqdm=True):
    """
    Convert list of images to pdf
    
    Args:
        images (list): list of image files
        pdf_file: file path of pdf to save 
        use_tqdm: if use tqdm progress bar
    
    Note:
        When prepare list of image files, do not need to judge the file type, 
        since this function will auto judge if the file can be read as image
    """
    assert pdf_file, 'arg pdf_file should not be empty'
    try:
        _ = Image
    except NameError as e:
        print('Error info: \n{}'.format(e))
        print('\nPlease import Image from PIL')
    
    if use_tqdm:
        for i, image in enumerate(tqdm(images, leave=False)):
            try:
                img = Image.open(image)
            except:
                continue
                
            if img.mode != "RGB":
                img = img.convert("RGB")
            
            if i == 0:
                img.save(pdf_file, 'PDF')
            else:
                img.save(pdf_file, 'PDF', append=True)
    else:
        for i, image in enumerate(images):
            try:
                img = Image.open(image)
            except:
                continue
                
            if img.mode != "RGB":
                img = img.convert("RGB")
            
            if i == 0:
                img.save(pdf_file, 'PDF')
            else:
                img.save(pdf_file, 'PDF', append=True)


def dir_to_pdf(dire, pdf_file='', use_tqdm=True):
    """
    Pack the use of images_to_pdf
    Auto file image files under given dire, then convert to a pdf file
    
    Args:
        dire: dire to search image files
        pdf_file: file path of pdf to save 
        use_tqdm: if use tqdm progress bar
    """
    pdf_file = pdf_file if pdf_file else Path(dire).name + '.pdf'
    # images = glob('{}/*'.format(dire))
    images = [os.path.join(dire, i) for i in os.listdir(dire)]
    images.sort()
    images_to_pdf(images, pdf_file, use_tqdm)


def get_sort_order_copymanga(name):
    name = Path(name).name
    dic = {'话': 0, '卷': 1, '番外': 2}
    key = re.findall('(c\d)(\w+)_(\d+) ', name)
    if not key:
        return ['c9', 9, '999']
    key = key[0]
    if not key[1] in dic.keys():
        return ['c9', 9, '999']
    else:
        return [key[0], dic[key[1]], key[2]]
    

def dirs_to_pdfs_copymanga_parallel(search_path='.', 
                                   book_name='new', 
                                   dire_pdf='pdfs',
                                   n_jobs=-1):
    """
    Special for copymanga, batch parallel convert dirs to pdfs
    
    Args:
        search_path: path to search sub_dire. default '.'
        book_name: comic book_name, used to rename the output pdf. 
            default 'pdfs'
        dire_pdf: dire to save the output pdf. default 'pdfs'
        n_jobs: n_jobs for Parallel. default -1
    """
    search_path = search_path.strip()
    if search_path != '.':
        finds = glob('*{}*'.format(search_path))
        if finds:
            search_path = finds[0]
            # print('    - find search_path: {}'.format(search_path))
        else:
            raise Exception('Cannot find path that match {}'.format(
                search_path))
    print('\nsearch path: {}'.format(search_path))
    
    dire_pdf = os.path.join(search_path, dire_pdf)
    if not os.path.exists(dire_pdf):
        os.makedirs(dire_pdf)
    
    # dires = [i for i in glob('{}/*'.format(search_path)) if os.path.isdir(i) and dire_pdf not in i]
    dires = [os.path.join(search_path, i) for i in os.listdir(search_path)]
    dires = [i for i in dires if os.path.isdir(i) and dire_pdf not in i]
    dires.sort(key=lambda x: get_sort_order_copymanga(x))
    print('\nfind dires:')
    for i in dires:
        print('    {}'.format(i))
    print()
    
    parallel = Parallel(n_jobs)
    _ = parallel(delayed(dir_to_pdf)(dire, 
            '{}/{} {} {}.pdf'.format(dire_pdf, book_name, 
                str(i+1).zfill(3), Path(dire).name.split()[1]), 
            False)
            for i, dire in enumerate(tqdm(dires)))


def main():
    print('\n{}'.format('-' * 50))
    print('Parallel convert image dires to pdfs for copymanga\n')
    
    search_path = input("path to search sub_dirs (default '.'): ").strip()
    book_name = input("comic name (default 'new'): ").strip()
    dire_pdf = input("output dire of pdfs (default 'pdfs'): ").strip()
    n_jobs = input("n_jobs for parallel (default '-1'): ").strip()
    
    if not search_path:
        search_path = '.'
    if not book_name:
        book_name = 'new'
    if not dire_pdf:
        dire_pdf = 'pdfs'
    if not n_jobs:
        n_jobs = -1
    
    dirs_to_pdfs_copymanga_parallel(search_path,
                                   book_name,
                                   dire_pdf,
                                   n_jobs)


if __name__ == '__main__':
    main()

    
    


# coding=utf-8
'''
Created on 
@author: hebing
'''
import re
from typing import List
import numpy as np
from ase.spacegroup.spacegroup import Spacegroup
from ase.spacegroup.spacegroup import parse_sitesym, SpacegroupValueError
from scipy.spatial.ckdtree import cKDTree

# from pymatgen.core.periodic_table  import Element


class MySpacegroup(Spacegroup):

    def __init__(self, spacegroup, setting=1, datafile=None):
        super().__init__(spacegroup, setting=setting, datafile=datafile)

    def equivalent_sites(self,
                         scaled_positions,
                         onduplicates='error',
                         symprec=1e-3):
        """Returns the scaled positions and all their equivalent sites.

        Parameters:

        scaled_positions: list | array
            List of non-equivalent sites given in unit cell coordinates.
        onduplicates : 'keep' | 'replace' | 'warn' | 'error'
            Action if `scaled_positions` contain symmetry-equivalent
            positions:

            'keep'
               ignore additional symmetry-equivalent positions
            'replace'
                replace
            'warn'
                like 'keep', but issue an UserWarning
            'error'
                raises a SpacegroupValueError

        symprec: float
            Minimum "distance" betweed two sites in scaled coordinates
            before they are counted as the same site.

        Returns:

        sites: array
            A NumPy array of equivalent sites.
        kinds: list
            A list of integer indices specifying which input site is
            equivalent to the corresponding returned site.

        Example:

        >>> from ase.spacegroup import Spacegroup
        >>> sg = Spacegroup(225)  # fcc
        >>> sites, kinds = sg.equivalent_sites([[0, 0, 0], [0.5, 0.0, 0.0]])
        >>> sites
        array([[ 0. ,  0. ,  0. ],
               [ 0. ,  0.5,  0.5],
               [ 0.5,  0. ,  0.5],
               [ 0.5,  0.5,  0. ],
               [ 0.5,  0. ,  0. ],
               [ 0. ,  0.5,  0. ],
               [ 0. ,  0. ,  0.5],
               [ 0.5,  0.5,  0.5]])
        >>> kinds
        [0, 0, 0, 0, 1, 1, 1, 1]
        """
        kinds = []
        sites = []
        scaled = np.array(scaled_positions, ndmin=2)
        for kind, pos in enumerate(scaled):
            for rot, trans in self.get_symop():
                site = np.mod(np.dot(rot, pos) + trans, 1.)
                if not sites:
                    sites.append(site)
                    kinds.append(kind)
                    continue
                t = site - sites
                mask = np.all(
                    (abs(t) < symprec) | (abs(abs(t) - 1.0) < symprec), axis=1)
                if np.any(mask):
                    ind = np.argwhere(mask)[0][0]
                    if kinds[ind] == kind:
                        pass
                    elif onduplicates == 'keep':
                        sites.append(site)
                        kinds.append(kind)
                    elif onduplicates == 'replace':
                        kinds[ind] = kind
                    elif onduplicates == 'warn':
                        warnings.warn('scaled_positions %d and %d '
                                      'are equivalent' % (kinds[ind], kind))
                    elif onduplicates == 'error':
                        raise SpacegroupValueError(
                            'scaled_positions %d and %d are equivalent' %
                            (kinds[ind], kind))
                    else:
                        raise SpacegroupValueError(
                            'Argument "onduplicates" must be one of: '
                            '"keep", "replace", "warn" or "error".')
                else:
                    sites.append(site)
                    kinds.append(kind)
        return np.array(sites), kinds


def spacegroup_from_data(no=None,
                         symbol=None,
                         setting=None,
                         centrosymmetric=None,
                         scaled_primitive_cell=None,
                         reciprocal_cell=None,
                         subtrans=None,
                         sitesym=None,
                         rotations=None,
                         translations=None,
                         datafile=None):
    """Manually create a new space group instance.  This might be
    useful when reading crystal data with its own spacegroup
    definitions."""
    if no is not None and setting is not None:
        spg = MySpacegroup(no, setting, datafile)
    elif symbol is not None:
        spg = MySpacegroup(symbol, setting, datafile)
    else:
        raise SpacegroupValueError('either *no* and *setting* '
                                   'or *symbol* must be given')
    if not isinstance(sitesym, list):
        raise TypeError('sitesym must be a list')

    have_sym = False
    if centrosymmetric is not None:
        spg._centrosymmetric = bool(centrosymmetric)
    if scaled_primitive_cell is not None:
        spg._scaled_primitive_cell = np.array(scaled_primitive_cell)
    if reciprocal_cell is not None:
        spg._reciprocal_cell = np.array(reciprocal_cell)
    if subtrans is not None:
        spg._subtrans = np.atleast_2d(subtrans)
        spg._nsubtrans = spg._subtrans.shape[0]
    if sitesym is not None:
        spg._rotations, spg._translations = parse_sitesym(sitesym)
        have_sym = True
    if rotations is not None:
        spg._rotations = np.atleast_3d(rotations)
        have_sym = True
    if translations is not None:
        spg._translations = np.atleast_2d(translations)
        have_sym = True
    if have_sym:
        if spg._rotations.shape[0] != spg._translations.shape[0]:
            raise SpacegroupValueError('inconsistent number of rotations and '
                                       'translations')
        spg._nsymop = spg._rotations.shape[0]
    return spg


class Site(object):
    '''
    '''

    def __init__(self, sitelabel=[], Pos=None, elementsoccupy=None):
        self._Id = 0
        self._SiteLabel = sitelabel
        self._SiteType = []
        self.IronType = 0
        self.radius = None
        self._FracPostion = Pos
        self._Elements_occupy = {}
        self._Elements_OxiValue = {}
        self.SetElementsOccupy(elementsoccupy)

    def SetSiteType(self, sitetype):
        if sitetype not in self._SiteType:
            self._SiteType.append(sitetype)

    def GetSiteType(self):
        return self._SiteType

    def SetSiteLabel(self, sitelabel):
        if sitelabel != self._SiteLabel:
            self._SiteLabel = self._SiteLabel + sitelabel

    def SetPostion(self, pos=None):
        self._FracPostion = pos

    def SetElementsOccupy(self, eleocu):
        for key in eleocu.keys():
            if key in self._Elements_occupy:
                # self._Elements_occupy[key]=self._Elements_occupy[key]+eleocu[key]
                pass
            else:
                self._Elements_occupy[key] = eleocu[key]

    def SetElementsOxiValue(self, eleoxi):
        for key in eleoxi.keys():
            self._Elements_OxiValue[key] = eleoxi[key]
            if eleoxi[key] < 0:
                self.IronType = -1
            elif eleoxi[key] > 0:
                self.IronType = 1
            else:
                self.IronType = 0

    def GetElementsOxiValue(self):
        return self._Elements_OxiValue

    def SetSiteId(self, siteid=0):
        self._Id = siteid

    def GetSiteId(self):
        return self._Id

    def GetSiteLabel(self):
        return self._SiteLabel

    def GetPosition(self):
        return self._FracPostion

    def GetElementsOccupy(self):
        return self._Elements_occupy

    def GetElements(self):
        return self._Elements_occupy.keys()

    def GetIronType(self):
        return self.IronType


class PeriodSite(Site):

    def __init__(self,
                 parent=None,
                 sitelabel=[],
                 sitetype=None,
                 Pos=None,
                 elementsoccupy=None):
        super(PeriodSite, self).__init__(sitelabel, sitetype, Pos,
                                         elementsoccupy)
        self.__ParentStruct = parent
        self.__ExpandCell = np.array([[0, 0, 0], [0, 0, 0]])

    def SetParentStructure(self, parent):
        self.__ParentStruct = parent

    def GetExpandCell(self):
        return self.__ExpandCell


class Bond(object):

    def __init__(self,
                 site0=None,
                 site1=None,
                 startpos=[0, 0, 0],
                 endpos=[0, 0, 0]):
        self._Site0 = site0
        self._Site1 = site1
        self._Positions = [startpos, endpos]

    def get_postions(self):
        return self._Positions

    def set_postions(self, value):
        self._Positions = value

    def del_postions(self):
        del self._Positions

    def get_site_0(self):
        return self._Site0

    def get_site_1(self):
        return self._Site1

    def set_site_0(self, value):
        self._Site0 = value

    def set_site_1(self, value):
        self._Site1 = value

    def del_site_0(self):
        del self._Site0

    def del_site_1(self):
        del self._Site1

    def SetSites(self, site0=None, site1=None):
        self._Site0 = site0
        self._Site1 = site1

    Site0 = property(get_site_0, set_site_0, del_site_0, "Site0's docstring")
    Site1 = property(get_site_1, set_site_1, del_site_1, "Site1's docstring")
    Positions = property(get_postions, set_postions, del_postions,
                         "Postions's docstring")


class Polyhedra(object):

    def __init__(self):
        self._VertexSites = []
        self._CentreSite = None
        self._VertexVectors = []

    def get_vertex_vectors(self):
        return self._VertexVectors

    def set_vertex_vectors(self, value):
        self._VertexVectors = value

    def del_vertex_vectors(self):
        del self._VertexVectors

    def get_centre_site(self):
        return self._CentreSite

    def set_centre_site(self, value):
        self._CentreSite = value

    def del_centre_site(self):
        del self._CentreSite

    def get_sites(self):
        return self._VertexSites

    def set_sites(self, value):
        self._VertexSites = value

    def del_sites(self):
        del self._VertexSites[:]

    def get_label(self):
        return self._CentreSite.GetSiteLabel()

    Label = property(get_label)
    VertexSites = property(get_sites, set_sites, del_sites,
                           "Sites's docstring")
    CentreSite = property(get_centre_site, set_centre_site, del_centre_site,
                          "CentreSites's docstring")
    VertexVectors = property(get_vertex_vectors, set_vertex_vectors,
                             del_vertex_vectors, "VertexVectors's docstring")


class Structure(object):

    def __init__(self):
        self.__abc = []
        self.__alphabetagama = []
        self.__positions = []
        self._initsites = []
        self._Sites: List[Site] = []
        self._SuperCellusites: List[Site] = []
        self._SuperCellusitesFracPos = []
        self._SuperCellusitesCartPos = []
        self.__SuperCellusitesCellIndex = []
        self.__spacegroupno = 1
        self.__ABC = []
        self.__equivviewsites = []
        self.__KDtree = None
        #       self.__ElementsSymbol={}
        self._Polyhedras = {}
        self._Bonds = {}
        self.__usitepos = []
        self._OxidationTable = {}
        self._atomsymbols = []

    def get_atomsymbols(self):
        return self._atomsymbols

    def get_polyhedras(self):
        return self._Polyhedras

    def set_polyhedras(self, value):
        self._Polyhedras = value

    def del_polyhedras(self):
        del self._Polyhedras

    def get_bonds(self):
        return self._Bonds

    def set_bonds(self, value):
        self._Bonds = value

    def del_bonds(self):
        del self._Bonds

    def Setabc(self, abc):
        self.__abc = abc

    def Getabc(self):
        return self.__abc

    def SetABC(self, ABC=None):
        self.__ABC = ABC

    def GetABC(self):
        return self.__ABC

    def GetSuperCellusitesFracPos(self):
        return self._SuperCellusitesFracPos

    def GetSuperCellusites(self):
        return self._SuperCellusites

    def GetSuperCellusitesCartPos(self):
        return self._SuperCellusitesCartPos

    def ExpandCell(self):
        indexpos = [-2.0, -1.0, 0.0, 1.0, 2.0]
        del self._SuperCellusites[:]
        del self._SuperCellusitesFracPos[:]
        del self._SuperCellusitesCartPos[:]
        for k in range(5):
            for j in range(5):
                for i in range(5):
                    for ste in self._Sites:
                        pos = ste.GetPosition()
                        stepos = [
                            pos[0] + indexpos[i], pos[1] + indexpos[j],
                            pos[2] + indexpos[k]
                        ]
                        self._SuperCellusitesFracPos.append(stepos)
                        cartpos = self.FracPosToCartPos(stepos)
                        self._SuperCellusitesCartPos.append(cartpos)
                        self._SuperCellusites.append(ste)
        self.__KDtree = cKDTree(self._SuperCellusitesCartPos)

    def GetNeighbors(self, centre, r):
        return self.__KDtree.query_ball_point(centre, r)

    def GetKNeighbors(self, centre, kn):
        ## kn: 1 - center self; 2 - center self and nn; 3 - center self, nn and 2nd-nn
        return self.__KDtree.query(centre, kn, p=2, n_jobs=-1)

    def FracPosToCartPos(self, FracPos):
        # pos=self.__ABC[0]*FracPos[0]+self.__ABC[1]*FracPos[1]+self.__ABC[2]*FracPos[2]
        pos = np.dot(FracPos, self.__ABC)
        return pos

    def AddBond(self,
                beginsite,
                endsite,
                beginshiftpos=np.array([0, 0, 0]),
                endshiftpos=np.array([0, 0, 0]),
                typeid=None):
        bond = Bond(beginsite, endsite)
        bond.Positions = [beginsite.GetPosition() + beginshiftpos, endsite.GetPosition() + endshiftpos]
        if typeid:
            if not typeid in self._Bonds:
                self._Bonds[typeid] = []
            self._Bonds[typeid].append(bond)
        else:
            if not beginsite.GetSiteId() in self._Bonds:
                self._Bonds[beginsite.GetSiteId()] = []
            self._Bonds[beginsite.GetSiteId()].append(bond)

    def CreateBonds(self, r):
        self._Bonds.clear()
        for ste in self._Sites:
            if ste.GetIronType() > 0:
                stepos = ste.GetPosition()
                cartpos = self.FracPosToCartPos(stepos)
                # neighbors=self.GetKNeighbors(cartpos, r=3)
                (distance, neighbors) = self.GetKNeighbors(cartpos, kn=12)
                if len(neighbors) > 1:
                    polyhedra = Polyhedra()
                    polyhedra.CentreSite = ste

                    for (i, index) in enumerate(neighbors[1:]):
                        site2 = self._SuperCellusites[index]
                        fracpos2 = self._SuperCellusitesFracPos[index]
                        if site2.GetIronType() < 0 and distance[i + 1] < r:
                            bond = Bond(ste, site2)
                            bond.Positions = [stepos, fracpos2]
                            if not ste.GetSiteId() in self._Bonds:
                                self._Bonds[ste.GetSiteId()] = []
                            self._Bonds[ste.GetSiteId()].append(bond)
                            polyhedra.VertexSites.append(site2)
                            polyhedra.VertexVectors.append(fracpos2 - stepos)
                        else:
                            break
                    self._Polyhedras[ste.GetSiteId()] = polyhedra

    def CreateAllBonds(self):
        self.CreateBonds(r=2.5)

    def SetSpaceGroupno(self, spacegroupid=1):
        self.__spacegroupno = spacegroupid

    def GetSpaceGroupno(self):
        return self.__spacegroupno

    def SetAlphaBetaGama(self, abg):
        self.__alphabetagama = abg

    def GetAlphaBetaGama(self):
        return self.__alphabetagama

    def AddSite(self, site=None):
        self._initsites.append(site)

    def AddUSite(self, site=None):
        self._Sites.append(site)

    def AddAtoms(self, sitelabel, sitetype, Pos, elementsoccupy, radius=None):
        index = len(self._Sites)
        site = Site(sitelabel, Pos, elementsoccupy)
        site.SetSiteType(sitetype)
        site.SetSiteId(index)
        if not radius:
            self.AddUSite(site)
        else:
            site.radius = radius
            self.AddUSite(site)

    def GetSites(self):
        return self._initsites

    def GetUSites(self):
        return self._Sites

    def GetASEAtoms(self, atoms=None):
        self.Setabc(atoms.get_cell_lengths_and_angles()[0:3])
        self.SetAlphaBetaGama(atoms.get_cell_lengths_and_angles()[3:6])
        self.__ABC = atoms.cell
        self.__positions = atoms.get_scaled_positions()
        strlabel = {}
        for index, ssite in enumerate(self.__positions):
            s = atoms.get_chemical_symbols()[index]
            if s in strlabel:
                strlabel[s] = strlabel[s] + 1
            else:
                strlabel[s] = 0
            m = re.search(r'([A-Z][a-z]?)', s)
            symbol = m.group(0)
            stelabel = symbol + str(strlabel[s])
            eleoccupy = {symbol: 1.0}
            self.AddSite(
                Site(sitelabel=stelabel,
                     sitetype=s,
                     Pos=ssite,
                     elementsoccupy=eleoccupy))
            self.__sites[index].SetSiteId(index)
        self.MergeDuplicateSite(0.001)
        self.ExpandCell()

    # self.CreateAllBonds()

    def GetPoscarFile(self, filename=None):
        self._filename = filename
        atoms = ase.io.read(filename)
        self.GetASEAtoms(atoms)

    def GetAseStructure(self, atoms=None):
        self.Setabc([
            atoms.info['_cell_length_a'], atoms.info['_cell_length_b'],
            atoms.info['_cell_length_c']
        ])
        self.SetAlphaBetaGama([
            atoms.info['_cell_angle_alpha'], atoms.info['_cell_angle_beta'],
            atoms.info['_cell_angle_gamma']
        ])
        # if '_symmetry_int_tables_number' in atoms.info:
        #     self.SetSpaceGroupno(atoms.info['_symmetry_int_tables_number'])
        # elif '_space_group_it_number' in atoms.info:
        #     self.SetSpaceGroupno(atoms.info['_space_group_it_number'])
        # else:
        #     print('can not find group no')
        #     return
        index = 0
        self.__ABC = atoms.cell
        postions = list(
            zip(atoms.info['_atom_site_fract_x'],
                atoms.info['_atom_site_fract_y'],
                atoms.info['_atom_site_fract_z']))

        no = None
        if '_space_group.it_number' in atoms.info:
            no = atoms.info['_space_group.it_number']
        elif '_space_group_it_number' in atoms.info:
            no = atoms.info['_space_group_it_number']
        elif '_symmetry_int_tables_number' in atoms.info:
            no = atoms.info['_symmetry_int_tables_number']
        if no:
            self.SetSpaceGroupno(no)

        symbolHM = None
        if '_space_group.Patterson_name_h-m' in atoms.info:
            symbolHM = atoms.info['_space_group.patterson_name_h-m']
        elif '_symmetry_space_group_name_h-m' in atoms.info:
            symbolHM = atoms.info['_symmetry_space_group_name_h-m']
        elif '_space_group_name_h-m_alt' in atoms.info:
            symbolHM = atoms.info['_space_group_name_h-m_alt']
        sitesym = None
        for name in [
                '_space_group_symop_operation_xyz',
                '_space_group_symop.operation_xyz',
                '_symmetry_equiv_pos_as_xyz'
        ]:
            if name in atoms.info:
                sitesym = atoms.info[name]
                break
            else:
                sitesym = None

        spgroup = 1
        if sitesym is not None:
            subtrans = [(0.0, 0.0, 0.0)]
            spgroup = spacegroup_from_data(no=no,
                                           symbol=symbolHM,
                                           setting=1,
                                           sitesym=sitesym,
                                           subtrans=subtrans)
        else:
            spgroup = MySpacegroup(self.GetSpaceGroupno())
        ssites, kinds = spgroup.equivalent_sites(postions, onduplicates='keep')
        self.__positions = ssites
        if '_atom_type_oxidation_number' in atoms.info:
            for i, atomtype in enumerate(atoms.info['_atom_type_symbol']):
                self._OxidationTable[atomtype] = int(
                    atoms.info['_atom_type_oxidation_number'][i])
        atomsymbols = set([])
        for ssite in ssites:
            s = atoms.info['_atom_site_type_symbol'][kinds[index]]
            m = re.search(r'([A-Z][a-z]?)', s)
            symbol = m.group(0)
            atomsymbols.add(symbol)
            stelabel = atoms.info['_atom_site_label'][kinds[index]]
            eleoccupy = {
                symbol: atoms.info['_atom_site_occupancy'][kinds[index]]
            }
            self.AddSite(
                Site(sitelabel=stelabel, Pos=ssite, elementsoccupy=eleoccupy))
            self._initsites[index].SetSiteType(s)
            self._initsites[index].SetSiteId(index)
            if len(self._OxidationTable) > 0:
                eleoxi = {symbol: self._OxidationTable[s]}
                self._initsites[index].SetElementsOxiValue(eleoxi)
            index = index + 1
        self._atomsymbols = [i for i in atomsymbols]
        self.MergeDuplicateSite(0.001)
        self.ExpandCell()
        self.CreateAllBonds()

    def MergeDuplicateSite(self, symprec=0.001):
        dupsites = [False] * len(self.__positions)
        sitelabels = []
        for index, sitepos in enumerate(self.__positions):
            if not dupsites[index]:
                self.__usitepos.append(self.__positions[index])
                # _site=Site(sitelabel=self._initsites[index].GetSiteLabel(),Pos=sitepos,elementsoccupy=self._initsites[index].GetElementsOccupy())
                self._Sites.append(self._initsites[index])
                sitelabels.append([self._initsites[index].GetSiteLabel()])
                self._Sites[-1].SetSiteId(len(self._Sites) - 1)
                t = sitepos - self.__positions
                mask = np.all(
                    (abs(t) < symprec) | (abs(abs(t) - 1.0) < symprec), axis=1)
                if np.any(mask):
                    inds = np.argwhere(mask).tolist()
                    if len(inds) > 1:
                        for dindex in inds:
                            dupsites[dindex[0]] = True
                            _occupy = self._initsites[
                                dindex[0]].GetElementsOccupy()
                            _sitelabel = self._initsites[
                                dindex[0]].GetSiteLabel()
                            _sitetype = self._initsites[
                                dindex[0]].GetSiteType()
                            if index != dindex[0] and (not _sitelabel
                                                       in sitelabels[-1]):
                                sitelabels[-1].append(_sitelabel)
                                for stype in _sitetype:
                                    # print(index)
                                    self._Sites[-1].SetSiteType(stype)
                                    m = re.search(r'([A-Z][a-z]?)', stype)
                                    symbol = m.group(0)
                                    if len(self._OxidationTable) > 0:
                                        eleoxi = {
                                            symbol: self._OxidationTable[stype]
                                        }
                                        self._Sites[-1].SetElementsOxiValue(
                                            eleoxi)
                                self._Sites[-1].SetElementsOccupy(_occupy)
                                self._Sites[-1].SetSiteLabel(_sitelabel)

    Bonds = property(get_bonds, set_bonds, del_bonds, "Bonds's docstring")
    Polyhedras = property(get_polyhedras, set_polyhedras, del_polyhedras,
                          "Polyhedras's docstring")



# -*- coding: utf-8 -*-
"""
author: Bowei Pu at 2023.03.01
version: 2024.01.15
function: Some useful tools
"""

from pybw.tricks.lazy_import import *


__author__ = 'Bowei Pu'
__version__ = '2024.01.15'
__maintainer__ = 'Bowei Pu'
__email__ = 'pubowei@foxmail.com'
__status__ = 'Beta'
__date__ = '2024-01-15'


def time_parser(sec):
    s = int(sec)
    m, sec = s // 60, s % 60
    h, mins = m // 60, m % 60
    day, hour = h // 24, h % 24
    return {'day': day, 'hour': hour, 'min': mins, 'sec': sec, 's': s}


def getsizeof(obj):
    """
    Get size of a object.
    """
    return sys.getsizeof(pickle.dumps(obj))


class EasyPickle():
    """
    Easy use of pickle.
    """
    def __init__(self):
        self._function = 'Easy use of pickle'
    
    @classmethod
    def dump(cls, obj, file_obj):
        pickle.dump(obj, file_obj)
    
    @classmethod
    def load(cls, file_obj):
        return pickle.load(file_obj)
    
    @classmethod
    def save(cls, obj, file):
        if not os.path.exists(Path(file).parent):
            os.makedirs(Path(file).parent)
        with open(file, 'wb') as f:
            pickle.dump(obj, f)
    
    @classmethod
    def read(cls, file):
        with open(file, 'rb') as f:
            return pickle.load(f)


class easypickle(EasyPickle):
    """
    Easy usage of pickle.
    """
    def __init__():
        EasyPickle.__init__(self)


class PickleDump(EasyPickle):
    """
    Compatible to old version programs.
    """
    def __init__():
        EasyPickle.__init__(self)


class DictDoc():
    """
    Pack dict to class for easy use.
    """
    def __init__(self, dic: dict={}):
        self._dic = dic
        self._generate_attributes()
    
    def __repr__(self):
        return '<DictDoc>'
    
    def __str__(self):
        return self.__repr__()

    def __getitem__(self, key):
        return self._dic[key]

    def _generate_attributes(self):
        """
        Only used when first time load dic.
        """
        for k, v in self._dic.items():
            setattr(self, k, v)
    
    @property
    def dict(self):
        """
        Generate vars dict up-to-date.
        """
        new_dic = vars(self).copy()
        _ = new_dic.pop('_dic')
        return new_dic
    
    @property
    def dic(self):
        return self.dict
    
    def as_dict(self):
        return self.dict
    
    def keys(self):
        return self.dict.keys()
    
    def values(self):
        return self.dict.values()
    
    def setattr(self, name, value):
        setattr(self, name, value)
    
    def add(self, name, value):
        self._dic[name] = value
        self.setattr(name, value)
    

class OsPath():
    """
    For some low version python that do not have pathlib.
    """
    def __init__(self, file):
        self.file = file
        
        self.path = self.abspath = os.path.abspath(self.file)
        self.parent, self.name = os.path.split(self.abspath)
        self.dirname, self.basename, self.filename = self.parent, self.name, self.name
        self.stem, self.suffix = os.path.splitext(self.name)
        self.path_stem = self.abs_stem = os.path.splitext(self.abspath)[0]
        
        self.str = self.string = self.abspath

    def __repr__(self):
        return self.abspath
        
    def __str__(self):
        return self.__repr__()
        
    def __fspath__(self):
        return self.abspath
    
    def absolute(self):
        return self.abspath
    
    def as_posix(self):
        return Path(self.abspath).as_posix()
    
    @property
    def posix(self):
        return self.as_posix()
    
    @staticmethod
    def get_posix(cls, path):
        return Path(path).as_posix()
    
    def join(self, other: str):
        return self.__class__(os.path.join(self.abspath, other))
    
    def joinpath(self, other: str):
        return self.join(other)
    
    @property
    def exists(self):
        return os.path.exists(self.abspath)
    
    def is_existes(self):
        return os.path.exists(self.abspath)
    
    def isfile(self):
        return os.path.isfile(self.abspath)
    
    def isdir(self):
        return os.path.isdir(self.abspath)
    
    def rename(self, new_name):
        return self.__class__(os.path.join(self.dirname, new_name))


def Progress(total, progress, symbol='-', arrow=False, time0=False):
    
    length = 30
    
    if isinstance(total, (int, float)):
        total = int(total)
    else:
        total = len(total)
    
    scale = total / length
    index = int(progress / scale)
    
    if len(symbol) > 1:
        symbol = symbol[0]
    if arrow:
        index = index - 1
        arrow = 1
    else:
        arrow = 0
    
    percent = progress / total
    
    if not time0 or not isinstance(time0, (int, float)):
        time_used = 0
        time_left = 0
    else:
        time_used = time.time() - time0
        time_total = time_used / percent
        time_left = time_total - time_used
        
        time_used = int(round(time_used))
        time_total = int(round(time_total))
        time_left = int(round(time_left))
    
    time_used = time_parser(time_used)
    
    bar = '{}%[{}{}{}] {}/{} [{}s<{}s]'.format(
        int(percent * 100), 
        symbol * index, 
        '>' * arrow,
        ' ' * (length - index), 
        progress, 
        total, 
        time_used, 
        time_left)
    
    print('\r{}'.format(bar), end='', file=sys.stderr, flush=True)
    
    
    
    
    
    


# -*- coding: utf-8 -*-
"""
author: Bowei Pu at 2023.04.10
version: 2024.01.15

Some small tools for scholar.
"""
import re
import ase.io
from pymatgen.core.structure import Structure
from pymatgen.io.ase import AseAtomsAdaptor


class StructureCif():
    """
    Parse structure file to get pymatgen Structure and ase Atoms.
    
    Parameters
    ----------
    file : str
        structure file, supporting cif and poscar.
    """
    def __init__(self, file):
        self.file = file
        self.type = 'cif' if self.file.endswith('cif') else 'poscar'
        
        self.atoms = self.get_ase_atoms()
        self.oxi_states = self.get_oxi_states()
        
        self.structure = self.get_pmg_stru_from_atoms()
        self.stru = self.structure
    
    def get_ase_atoms(self):
        if self.type == 'cif':
            return ase.io.read(self.file, format='cif', store_tags=True)
        elif self.type == 'poscar':
            return ase.io.read(self.file, format='vasp')
    
    def get_oxi_states(self):
        """Require atoms read from cif with oxi states"""
        symbols = self.atoms.info['_atom_type_symbol']
        symbols = [ re.findall('[A-Z][a-z]*', i)[0] for i in symbols]
        oxis = self.atoms.info['_atom_type_oxidation_number']
        oxi_states = {s: o for s, o in zip(symbols, oxis)}
        return oxi_states
    
    def get_pmg_stru(self):
        try:
            stru = Structure.from_file(self.file)
        except:
            stru = AseAtomsAdaptor.get_structure(self.atoms)
            stru.add_oxidation_state_by_element(self.oxi_states)
        return stru
    
    def get_pmg_stru_from_atoms(self):
        stru = AseAtomsAdaptor.get_structure(self.atoms)
        stru.add_oxidation_state_by_element(self.oxi_states)
        return stru
        
        
        


# -*- coding: utf-8 -*-
"""
Modified version of ccnb main.py
"""
import os, time, shutil, re, argparse
from glob import glob
from pathlib import Path
from pymatgen.analysis.bond_valence import BVAnalyzer
from pymatgen.core.structure import Structure as Structure_pmg

from ccnb import cal_channel_cavd, get_bvse, get_migration_networks_voids


def get_voids(filename_CIF,
              energythreshold,
              moveion,
              mergecluster=True,
              clusterradii=0.75, 
              filename_CAVD=None, 
              filename_BVSE=None):
    """
    Get migration networks voids base on cal result of cavd and bvse
    
    Args:
        filename_CIF: cif filename. Used to give crastal structure and 
                      auto find filename_CAVD and filename_BVSE
        energythreshold: energythreshold
        moveion: move ion
    
    Returns:
        get_migration_networks_voids()
    """
    filename_CIF = Path(filename_CIF)
    if not filename_CAVD:
        filename_CAVD = str(filename_CIF.parent.joinpath(
            filename_CIF.stem)) + '.net'
    if not filename_BVSE:
        filename_BVSE = str(filename_CIF.parent.joinpath(
            filename_CIF.stem)) + '_BVSE.npy'
    
    if not Path(filename_CAVD).exists():
        print('{} file no exist,\n please cavd calculate first'.format(
            filename_CAVD))
        return
    if not Path(filename_BVSE).exists():
        print('{} file no exist,\n please bvse calculate first'.format(
            filename_BVSE))
        return
    
    mn = get_migration_networks_voids(str(filename_CIF),
                                      filename_BVSE,
                                      filename_CAVD,
                                      energythreshold,
                                      moveion=moveion,
                                      mergecluster=mergecluster,
                                      clusterradii=clusterradii)
    return mn


def get_argparse():
    """
    Old version program's function main
    Now it is changed to obtain argparse
    
    Returns:
        argparse.ArgumentParser()
    """
    parser = argparse.ArgumentParser(description='ccnb (mod ver. by pbw)')
    
    parser.add_argument('-f',
                        '-file',
                        '--structure_file',
                        dest='struct_file',
                        type=str,
                        default='',
                        required=False,
                        help='结构文件名 (compatible cif, POSCAR)')
    parser.add_argument('-i',
                        '-ion', 
                        '--move_ion',
                        dest='ion',
                        type=str,
                        default='Li',
                        required=False,
                        help='迁移离子类型')
    parser.add_argument('-c',
                        '-cal', 
                        '-mode', 
                        '-cal_type', 
                        '--calculation_type',
                        dest='cal_type',
                        type=str,
                        default='bvse',
                        choices=['bvse', 'cavd', 'find_voids', 'all'],
                        required=False,
                        help='计算类型选择')
    parser.add_argument('-p', 
                        '-pack',
                        '--pack',
                        dest='pack',
                        type=str,
                        default='False',
                        help='是否打包整理计算结果文件')
    
    bvse = parser.add_argument_group('bvse')
    bvse.add_argument('-v',
                      '-valence', 
                      '--valence',
                      type=float,
                      default=1,
                      dest='valence',
                      help='迁移离子化合价')
    bvse.add_argument('-r',
                      '-res', 
                      '-resolution', 
                      '--resolution',
                      type=float,
                      default=0.5,
                      dest='resolution',
                      help='计算区域分辨率（单位埃）')
    
    cavd = parser.add_argument_group('cavd')
    cavd.add_argument('--ntol',
                      dest='ntol',
                      type=float,
                      default=0.02,
                      help='计算容限')
    cavd.add_argument('--radius_flag',
                      dest='rad_flag',
                      type=str,
                      default='True',
                      help='几何分析时是否考虑离子半径')
    cavd.add_argument('-l',
                      '--lower',
                      dest='lower',
                      type=float,
                      default=0.5,
                      help='通道大小下限值(单位埃)')
    cavd.add_argument('-u',
                      '--upper',
                      dest='upper',
                      type=float,
                      default=1.0,
                      help='通道大小上限值(单位埃)')

    find_voids = parser.add_argument_group('find_voids')
    find_voids.add_argument('-e', 
                            '-energy', 
                            '--energy',
                            dest='energy',
                            type=float,
                            default=0.5,
                            help='能量筛选上限')
    find_voids.add_argument('-cluster', 
                            '--cluster',
                            dest='cluster',
                            type=str,
                            default='True',
                            help='是否进行合并团簇间隙点')
    find_voids.add_argument('-cr',
                            '-c_rad', 
                            '-clu_rad', 
                            '--cluster_radii',
                            dest='cluster_radii',
                            type=float,
                            default=0.75,
                            help='合并团簇范围半径')
    """
    args = parser.parse_args()
    if args.cal_type == 'cavd':
        RT = cal_channel_cavd(args.struct_file,
                              migrant=args.ion,
                              ntol=args.ntol,
                              rad_flag=args.rad_flag,
                              lower=args.lower,
                              upper=args.upper,
                              rad_dict=None)
        print(RT)
    if args.cal_type == 'bvse':
        barrier = get_bvse(args.struct_file,
                           moveion=args.ion,
                           valenceofmoveion=args.valence,
                           resolution=args.resolution)
        print(barrier)
    if args.cal_type == 'find_voids':
        mn = get_voids(filename_CIF=args.struct_file,
                       energythreshold=args.energy,
                       moveion=args.ion,
                       mergecluster=args.cluster,
                       clusterradii=args.cluster_radii)
        if mn:
            print(mn)
        else:
            print('Calculation failed!!!')
    """
    return parser


class CCNBCal():
    """
    Execute ccnb cal
    
    Attributes:
        (Including all attributes for cavd, bvse, ccnb)
        cal_type: 'cavd', 'bvse', 'find_voids', 'all'
        cal_now: whether cal immediately
        pack: whether pack result files after cal
        args: easy way to get all related args
    """
    def __init__(self, 
                 file='', 
                 ion='Li', 
                 valence=1, 
                 resolution=0.5, 
                 lower=0.5, 
                 upper=1.0, 
                 ntol=0.02, 
                 rad_flag=True, 
                 cutoff_energy=0.5, 
                 cluster=True, 
                 cluster_radii=0.75, 
                 cal_type='all', 
                 cal_now=True, 
                 pack=False, 
                 args=None):
        self.args = args
        if self.args:
            self.cal_type = self.args.cal_type
            # print('\npara pack of args: {}\n'.format(self.args.pack))
            self.pack = self.args.pack
            self.pack = False if str(self.pack) == 'False' else True
            
            self.file_origin = self.args.struct_file
            self.ion = self.args.ion
            self.valence = self.args.valence
            self.resolution = self.args.resolution
            
            self.lower = self.args.lower
            self.upper = self.args.upper
            self.ntol = self.args.ntol
            self.rad_flag = self.args.rad_flag
            
            self.cutoff_energy = self.args.energy
            self.cluster = self.args.cluster
            self.cluster_radii = self.args.cluster_radii
        else:
            self.cal_type = cal_type
            self.pack = pack
            
            self.file_origin = file
            self.ion = ion
            self.valence = valence
            self.resolution = resolution
            
            self.lower = lower
            self.upper = upper
            self.ntol = ntol
            self.rad_flag = rad_flag
            
            self.cutoff_energy = cutoff_energy
            self.cluster = cluster
            self.cluster_radii = cluster_radii
        
        self.prepare()
        
        if cal_now:
            self.cal_all()
        
    def prepare(self):
        self.task_state = {'continue': False, 'cavd': False, 
                           'bvse': False, 'find_voids': False}
        has_prefix = re.findall('^ccnb_[\d-]+--', Path(self.file_origin).name)
        if has_prefix:
            self.task_state['continue'] = True
            self.prefix = has_prefix[0]
            self.file = self.file_origin
            
            self.file_cavd = os.path.splitext(self.file)[0] + '.net'
            self.file_bvse = os.path.splitext(self.file)[0] + '_BVSE.npy'
            if os.path.exists(self.file_cavd):
                self.task_state['cavd'] = True
            if os.path.exists(self.file_bvse):
                self.task_state['bvse'] = True
        else:
            self.prefix = 'ccnb_' + time.strftime('%y%m%d%H%M') + '--'
            self.file = Path(self.file_origin).parent.joinpath(
                self.prefix + Path(self.file_origin).name).__str__()
            shutil.copy2(self.file_origin, self.file)
            
            stru = Structure_pmg.from_file(self.file)
            state = {'has_oxi_state': True, 'is_ordered': True}
            if not all(hasattr(ele, 'oxi_state') for site in stru 
                       for ele in site.species):
                state['has_oxi_state'] = False
                stru.remove_oxidation_states()
                stru = BVAnalyzer().get_oxi_state_decorated_structure(stru)
            if not stru.is_ordered:
                state['is_ordered'] = False
                from pymatgen.transformations.standard_transformations import \
                    OrderDisorderedStructureTransformation as ODST
                try:
                    stru = ODST().apply_transformation(stru)
                except:
                    raise Exception('\nThis disorder structure can not be '
                                    'transformed to ordered structure\n'
                                    'Please check the cif or try to order it '
                                    'throungh other ways\n')
            self.file = os.path.splitext(self.file)[0]
            if not state['has_oxi_state']:
                self.file = self.file + '-add_oxi_state'
            if not state['is_ordered']:
                self.file = self.file + '-ordered'
            self.file = self.file + '.cif'
            stru.to(fmt='cif', filename=self.file)
    
    def pack_files(self):
        dire = Path(self.file_origin).parent.joinpath(Path(self.file).stem)
        files = glob(r'{}/{}*'.format(Path(self.file).parent, self.prefix))
        os.makedirs(dire)
        for file in files:
            new = Path(file).name.lstrip(self.prefix)
            shutil.move(file, Path(dire).joinpath(new))
    
    def cal_cavd(self):
        RT = cal_channel_cavd(self.file,
                              migrant=self.ion,
                              ntol=self.ntol,
                              rad_flag=self.rad_flag,
                              lower=self.lower,
                              upper=self.upper,
                              rad_dict=None)
        print(RT)
    
    def cal_bvse(self):
        barrier = get_bvse(self.file,
                           moveion=self.ion,
                           valenceofmoveion=self.valence,
                           resolution=self.resolution)
        print(barrier)
    
    def cal_voids(self):
        mn = get_voids(filename_CIF=self.file,
                       energythreshold=self.cutoff_energy,
                       moveion=self.ion,
                       mergecluster=self.cluster,
                       clusterradii=self.cluster_radii, 
                       filename_CAVD=None, 
                       filename_BVSE=None)
        if mn:
            print(mn)
        else:
            print('\nCaution: Calculation failed\n')
    
    def cal_all(self):
        if self.cal_type.lower() == 'cavd':
            self.cal_cavd()
        elif self.cal_type.lower() == 'bvse':
            self.cal_bvse()
        elif self.cal_type.lower() == 'find_voids':
            self.cal_voids()
        elif self.cal_type.lower() in ['all']:
            if not self.task_state['cavd']:
                print('\n\n{0}\n[Step 01] cavd\n{0}\n'.format('-'*60))
                self.cal_cavd()
            if not self.task_state['bvse']:
                print('\n\n{0}\n[Step 02] bvse\n{0}\n'.format('-'*60))
                self.cal_bvse()
            print('\n\n{0}\n[Step 03] find_voids\n{0}\n'.format('-'*60))
            self.cal_voids()
        if self.pack:
            self.pack_files()


class Parameter():
    def __init__(self, dic):
        self.dic = dic
        
        for k, v in self.dic.items():
            setattr(self, k, v)
    
    def __repr__(self):
        return '<Parameter Storage>'


def receive_input():
    print('\n ****************************\n'
            ' *                          *\n'
            ' *         CCNB Mod         *\n'
            ' *      ver. 23.3.15.1      *\n'
            ' *                          *\n'
            ' ****************************\n'
            ' [Special for A-SOUL AVA]\n')
            
    # print('\n+--------------------------+\n'
            # '|                          |\n'
            # '|         CCNB Mod         |\n'
            # '|      ver. 23.3.15.1      |\n'
            # '|                          |\n'
            # '+--------------------------+\n'
            # '[Special for A-SOUL AVA]\n')

    print('\nParameters for cal control')
    print('    [cal_type: cavd; bvse; find_voids; all]')
    cal_type = input('    cal_type [all] = ').strip().lower()
    pack = input('    pack [False] = ').strip()
    file = input('\n    Structure file [first cif file] = ').strip()
    
    print('\nParameters for BVSE')
    ion = input("    ion ['Li'] = ").strip()
    valence = input('    valence [1] = ').strip()
    resolution = input('    resolution [0.5] = ').strip()
    print('\nParameters for CAVD')
    lower = input('    lower [0.5] = ').strip()
    upper = input('    upper [1.0] = ').strip()
    ntol = input('    ntol [0.02] = ').strip()
    rad_flag = input('    rad_flag [True] = ').strip()
    print('\nParameters for get_networks_voids')
    cutoff_energy = input('    cutoff_energy [0.5] = ').strip()
    cluster = input('    cluster [True] = ').strip()
    cluster_radii = input('    cluster_radii [0.75] = ').strip()
    print('\n{}\n'.format('-' * 60))
    
    paras = {}
    paras['cal_type'] = cal_type if cal_type else 'all'
    paras['pack'] = False if pack.strip() == 'False' else True
    paras['file'] = (file if file else 
                     (glob('*.cif')[0] if glob('*.cif') else ''))
    paras['struct_file'] = paras['file']
    paras['ion'] = ion if ion else 'Li'
    paras['valence'] = float(valence) if valence else 1
    paras['resolution'] = float(resolution) if resolution else 0.5
    paras['lower'] = float(lower) if lower else 0.5
    paras['upper'] = float(upper) if upper else 1.0
    paras['ntol'] = float(ntol) if ntol else 0.02
    paras['rad_flag'] = False if rad_flag.strip() == 'False' else True
    paras['cutoff_energy'] = float(cutoff_energy) if cutoff_energy else 0.5
    paras['energy'] = paras['cutoff_energy']
    paras['cluster'] = False if cluster.strip() == 'False' else True
    paras['cluster_radii'] = float(cluster_radii) if cluster_radii else 0.8
    print('\nParameters: {}\n'.format(paras))
    return Parameter(paras)


def main():
    parser = get_argparse()
    if not parser.parse_args().struct_file:
        paras = receive_input()
        ccnb_cal = CCNBCal(args=paras)
    else:
        args = parser.parse_args()
        print('\nParameters: {}\n'.format(args))
        ccnb_cal = CCNBCal(args=args)


if __name__ == '__main__':
    main()
    


# -*- coding: utf-8 -*-
# function: lazy import


## ------ System ------
import os
import sys
import gc
import copy
import shutil
import time
# from datetime import datetime
from glob import glob
from pathlib import Path

try:
    from dotted_dict import DottedDict
except:
    pass


## ------ System Pro ------
import _pickle as pickle
try:
    from tqdm import tqdm
except:
    pass

try:
    from joblib import Parallel, delayed
except:
    pass

# try:
    # from func_timeout import func_set_timeout, func_timeout
# except:
    # pass


## ------ Data Analysis ------
import re
import random
import json
import yaml

try:
    from prettytable import PrettyTable
except:
    pass

import numpy as np
try:
    import pandas as pd
except:
    pass
try:
    import polars as pl
except:
    pass

import matplotlib.pyplot as plt
plt.rcParams['font.family'] = ['Arial']
# plt.rcParams['font.family'] = ['sans-serif']
# plt.rcParams['font.sans-serif'] = ['SimHei']
# plt.rcParams['axes.unicode_minus'] = False
plt.ion()
# import scienceplots
plt.style.use('ggplot')

try:
    import seaborn as sns
except:
    pass


## ------ Scientific Calculation ------
import math
from decimal import Decimal


## ------ Audio Video ------
# from moviepy.editor import VideoFileClip, AudioFileClip


## ------ pybw ------
from pybw.core import *





# -*- coding: utf-8 -*-
'''
Origin code info:
    Created on 2017
    author: hebing

Modify info:
    modify author: Bowei Pu
    version: 2022.08.17
    function:
        Consider mobile ion when cal bv
'''
import os
import numpy as np
from scipy.special import erfc
from scipy.ndimage import measurements
import scipy.ndimage
import math
from struct import pack
from tqdm import tqdm


def loadbvparam(filename):
    with open(filename, 'r') as fp:
        lines = fp.readlines()
    bvmpara = {}
    for line in lines:
        varstr = line.split('\t')
        key = varstr[0] + varstr[1] + varstr[2] + varstr[3]
        if key in bvmpara:
            bvmpara[key].append([float(varstr[4]), float(varstr[5])])
        else:
            bvmpara[key] = [[float(varstr[4]), float(varstr[5])]]
    return bvmpara


def loadBVSEparam(filename):
    with open(filename, 'r') as fp:
        lines = fp.readlines()
    BVSEparam = {}
    for line in lines[1:]:
        line = line.strip('\n')
        varstr = line.split('\t')
        key = varstr[0] + varstr[1] + varstr[2] + varstr[3]
        if key in BVSEparam:
            BVSEparam[key].append([
                float(varstr[4]),
                float(varstr[5]),
                float(varstr[6]),
                float(varstr[7]),
                float(varstr[8]),
                float(varstr[9])
            ])
        else:
            BVSEparam[key] = [[
                float(varstr[4]),
                float(varstr[5]),
                float(varstr[6]),
                float(varstr[7]),
                float(varstr[8]),
                float(varstr[9])
            ]]
    return BVSEparam


def loadElementsparam(filename):
    with open(filename, 'r') as fp:
        lines = fp.readlines()
    Elementsparam = {}
    for line in lines:
        line = line.strip('\n')
        varstr = line.split('\t')
        key = varstr[1] + varstr[2]
        Elementsparam[key] = [
            float(varstr[0]),
            float(varstr[3]),
            float(varstr[4]),
            float(varstr[5]),
            float(varstr[6]),
            float(varstr[7]),
            float(varstr[8]),
            float(varstr[9])
        ]
    return Elementsparam


class BVAnalysis(object):
    '''
    for bond valence method analyze ion channel
    '''

    def __init__(self, struc=None, size=[50, 50, 50]):
        '''
        Constructor
        '''
        self._Struct = struc
        self._Size = [50, 50, 50]
        self.__Data = {}
        self.__Max = {}
        self.__Min = {}
        self.Ea = {}
        self._MoveIon = 'Li'
        self._poss = None
        self.ValenceOfMoveIon = 1
        self.stop = False
        module_dir = os.path.dirname(os.path.abspath(__file__))

        self._BVparam = loadbvparam(os.path.join(module_dir, 'bvmparam.dat'))
        self._BVSEparam = loadBVSEparam(os.path.join(module_dir, 'bvse.dat'))
        self._Elementsparam = loadElementsparam(os.path.join(module_dir, 'elements.dat'))

    def get_data(self):
        return self.__Data

    def get_max(self):
        return self.__Max

    def get_min(self):
        return self.__Min

    def set_data(self, value):
        self.__Data = value

    def set_max(self, value):
        self.__Max = value

    def set_min(self, value):
        self.__Min = value

    def del_data(self):
        del self.__Data

    def del_max(self):
        del self.__Max

    def del_min(self):
        del self.__Min

    @property
    def BVSMin(self):
        return self.__Min['BVS']

    @property
    def BVSEMin(self):
        return self.__Min['BVSE']

    @property
    def BVELMin(self):
        return self.__Min['BVEL']

    @property
    def BVSMax(self):
        return self.__Max['BVS']

    @property
    def BVSEMax(self):
        return self.__Max['BVSE']

    @property
    def BVELMax(self):
        return self.__Max['BVEL']

    @property
    def BVSEEa(self):
        return self.Ea['BVSE']

    @property
    def BVELEa(self):
        return self.Ea['BVEL']

    def SetStructure(self, struc):
        self._Struct = struc

    def SetMoveIon(self, moveion):
        self._MoveIon = moveion

    def SetResolution(self, size):
        self._Size = size

    def SetLengthResolution(self, reslen=0.1):
        abc = self._Struct.Getabc()
        self._reslen = reslen
        self._Size[0] = int(abc[0] / reslen) + 1
        self._Size[1] = int(abc[1] / reslen) + 1
        self._Size[2] = int(abc[2] / reslen) + 1

    def SaveBVSEData(self, filename):
        self.SaveBinaryData('{}_BVSELi.pgrid'.format(filename), self.__Data['BVSE'], 'BVSE')
        np.save('{}_BVSELi.npy'.format(filename), self.__Data['BVSE'])

    def SaveBVELData(self, filename):
        self.SaveBinaryData('{}_BVELLi.pgrid'.format(filename), self.__Data['BVEL'], 'BVEL')
        # np.save('{}_BVELLi.npy'.format(filename), self.__Data['BVEL'])

    def SaveBVSData(self, filename):
        self.SaveBinaryData('{}_BVSLi.pgrid'.format(filename), self.__Data['BVS'], 'BVS')
        # np.save('{}_BVSLi.npy'.format(filename), self.__Data['BVS'])

    def SaveData(self, filename, data, datatype=' '):
        fp = open(filename, 'w')
        fp.write(datatype + '\n')
        alpha = self._Struct.GetAlphaBetaGama()
        abc = self._Struct.Getabc()
        fp.write(str(abc[0]) + ' ' + str(abc[1]) + ' ' + str(abc[2]) + ' ')
        fp.write(str(alpha[0]) + ' ' + str(alpha[1]) + ' ' + str(alpha[2]) + '\n')
        fp.write(str(self._Size[0]) + ' ' + str(self._Size[1]) + ' ' + str(self._Size[2]) + '\n ')
        for i in range(self._Size[0]):
            for j in range(self._Size[1]):
                for k in range(self._Size[2]):
                    fp.write(str(data[i][j][k]) + '\n')
        fp.close()

    def SaveBinaryData(self, filename, data, datatype=' '):
        latticeparam = self._Struct.Getabc() + self._Struct.GetAlphaBetaGama()
        with open(filename, 'wb') as outf:
            outf.write(pack('4i', 3, 0, 0, 0))  # file format version
            outf.write(pack(
                '80s',
                datatype.ljust(80).encode('utf-8')))  # Title 80 characters
            outf.write(pack('i', 1))  # gType 0 for *.ggrid, 1 for *.pgrid
            outf.write(pack('i', 0))  # fType 0
            outf.write(pack('i', 1))  # nVal dummy
            outf.write(pack('i', 3))  # dimension
            outf.write(pack('3i', *self._Size))  # numbers of voxels
            outf.write(pack('i', data.size))  # Total number of voxels
            outf.write(pack('6f', *latticeparam))  # lattice parameters
            for k in range(self._Size[2]):
                for j in range(self._Size[1]):
                    for i in range(self._Size[0]):
                        outf.write(pack('f', data[i][j][k]))

    def ReadData(self, filename, data, datatype=' '):
        fp = open(filename, 'r')
        datatype = fp.readline()
        alpha = self._Struct.GetAlphaBetaGama()
        abc = self._Struct.Getabc()
        abc = fp.readline().split()
        alpha = fp.readline().split()
        strs = fp.readline().split()
        size = [int(i) for i in strs]
        if datatype == 'bvs':
            self.__Data['BVS'] = np.zeros(size)
            data = self.__Data['BVS']
        else:
            self.__Data['BVSE'] = np.zeros(size)
            data = self.__Data['BVSE']
        for i in range(size[2]):
            for j in range(size[1]):
                for k in range(size[0]):
                    strs = fp.readline().split()
                    data[k][j][i] = int(strs[0])
        fp.close()

    def GetBVSData(self):
        return self.__Data['BVS']

    def GetBVSEData(self):
        return self.__Data['BVSE']

    def GetPosSet(self):
        return self._poss

    def CaluGlobalIndex(self):
        self.stop = False
        self._Size.reverse()
        centrepos = np.zeros(self._Size + [3])
        self._Data = np.zeros(self._Size, dtype=np.double)
        for k in range(self._Size[0]):
            for j in range(self._Size[1]):
                for i in range(self._Size[2]):
                    centrepos[k][j][i][2] = k / (self._Size[0] - 1.0)
                    centrepos[k][j][i][1] = j / (self._Size[1] - 1.0)
                    centrepos[k][j][i][0] = i / (self._Size[2] - 1.0)

        self._poss = self._Struct.FracPosToCartPos(centrepos)
        (distance, neighborsindex) = self._Struct.GetKNeighbors(self._poss, kn=8)

        site2atoms = None
        for k in range(self._Size[0]):
            for j in range(self._Size[1]):
                for i in range(self._Size[2]):
                    if self.stop:
                        return
                    
                    for dindex, index in enumerate(neighborsindex[k][j][i]):
                        site2 = self._Struct.GetSuperCellusites()[index]
                        if site2.GetIronType() * self.ValenceOfMoveIon < 0:
                            for key in site2.GetElementsOccupy():
                                if key != 'Vac':
                                    site2atoms = key
                            if self.ValenceOfMoveIon > 0:
                                key = "".join([self._MoveIon, str(self.ValenceOfMoveIon), site2atoms, str(site2.GetElementsOxiValue()[site2atoms])])
                            else:
                                key = "".join([site2atoms, str(site2.GetElementsOxiValue()[site2atoms]), self._MoveIon, str(self.ValenceOfMoveIon)])
                            if key in self._BVparam:
                                (r, b) = self._BVparam[key][0]
                                bv = np.exp((r - distance[k][j][i][dindex]) / b)
                                self.__Data['BVS'][k][j][i] = self.__Data['BVS'][k][j][i] + bv
                            else:
                                print('Not Find bvs param' + key)
                        else:
                            pass
                            # print(site2.GetSiteLabel())
    
    def CaluBVSE(self, ProgressCall, Rmin_moveion=3):
        self.Rmin_moveion = Rmin_moveion
        
        if len(self._Struct._OxidationTable) == 0:
            raise Exception('can not calculate BV and Ea without atom oxi info')
        
        self.stop = False
        
        centrepos = np.zeros(self._Size + [3])
        self.__Data['BVS'] = np.zeros(self._Size, dtype=np.double)
        self.__Data['BVSE'] = np.zeros(self._Size, dtype=np.double)
        self.__Data['BVEL'] = np.zeros(self._Size, dtype=np.double)
        
        for k in range(self._Size[2]):
            for j in range(self._Size[1]):
                for i in range(self._Size[0]):
                    centrepos[i][j][k][2] = k / (self._Size[2] - 1.0)
                    centrepos[i][j][k][1] = j / (self._Size[1] - 1.0)
                    centrepos[i][j][k][0] = i / (self._Size[0] - 1.0)
        
        if ProgressCall:
            ProgressCall(0)
        
        self._poss = self._Struct.FracPosToCartPos(centrepos)

        atomsq = {}
        for atom in self._Struct._atomsymbols:
            atomsq[atom] = 0
            for site in self._Struct._Sites:
                if atom in site.GetElements():
                    key = atom + str(site.GetElementsOxiValue()[atom])
                    atomsq[atom] = atomsq[atom] + site.GetElementsOxiValue()[atom] * site.GetElementsOccupy()[atom] / math.sqrt(self._Elementsparam[key][3])
        
        qsumanion = 0
        qsumcation = 0
        for (atom, value) in atomsq.items():
            if value > 0:
                qsumcation = qsumcation + value
            elif value < 0:
                qsumanion = qsumanion + value
            else:
                qsumanion = 0
                qsumcation = 0
        
        qsum = 0.0
        if self.ValenceOfMoveIon > 0:
            qsum = -qsumanion / qsumcation
        else:
            qsum = -qsumcation / qsumanion
        
        key1 = self._MoveIon + str(self.ValenceOfMoveIon)
        qm1 = self.ValenceOfMoveIon / math.sqrt(self._Elementsparam[key1][3])
        rm1 = self._Elementsparam[key1][6]
        
        ## tqdm color: '#00ff00', 'black', 'red', 'green', 'yellow', 'blue', 'magenta', 'cyan', 'white'
        for i in tqdm(range(self._Size[0]), colour='cyan', desc=None):
            (distance, neighborsindex) = self._Struct.GetKNeighbors(self._poss[i], kn=128)
            
            if ProgressCall:
                ProgressCall(10 + i * 90 / (self._Size[0] - 1))
            
            for j in tqdm(range(self._Size[1]), colour=None, leave=False):
                
                ## Single calculation is too fast, not need to use k loop tqdm
                # for k in tqdm(range(self._Size[2]), colour='black', leave=False):
                for k in range(self._Size[2]):
                    
                    if self.stop:
                        return False
                    
                    bvsdata = 0.0
                    data = 0.0
                    cdata = 0.0
                    bvelcdata = 0.0
                    N = 0
                    D0 = 0.0
                    Rcutoff = 10.0
                    alpha = 0.0
                    occupyvalue = 0.0
                    
                    list_cdata_Li = []
                    list_bvelcdata_Li = []
                    
                    break_this_point = False
                    exclude_NN_Li = False
                    
                    for dindex, index in enumerate(neighborsindex[j][k]):
                        if break_this_point:
                            break
                        
                        ## site2 is the neighbor of attempt_Li+
                        site2 = self._Struct._SuperCellusites[index]
                        
                        if site2.GetIronType() * self.ValenceOfMoveIon < 0:
                            for (ssymbol, occupyvalue) in site2.GetElementsOccupy().items():
                                if ssymbol != 'Vac':
                                    site2oxi = site2.GetElementsOxiValue()[ssymbol]
                                    if self.ValenceOfMoveIon > 0:
                                        key = "".join([self._MoveIon, str(self.ValenceOfMoveIon), ssymbol, str(site2oxi)])
                                    else:
                                        key = "".join([ssymbol, str(site2oxi), self._MoveIon, str(self.ValenceOfMoveIon)])
                                    if (key in self._BVSEparam) and (key in self._BVparam):
                                        (Nc, r0, Rcutoff, D0, Rmin, alpha) = self._BVSEparam[key][0]
                                        (r, b) = self._BVparam[key][0]
                                        Rcutoff = 10.0
                                        if distance[j][k][dindex] <= Rcutoff:
                                            # bv=np.exp((r0-distance[k][j][i][dindex])*alpha)
                                            bvs = np.exp((r - distance[j][k][dindex]) / b)
                                            smin = np.exp((Rmin - distance[j][k][dindex]) * alpha)
                                            en_s = np.exp((Rmin - Rcutoff) * alpha)
                                            bvsdata = bvsdata + bvs  #*occupyvalue-((en_s-1)**2-1)
                                            data = data + ((smin - 1)**2 - 1) * occupyvalue
                                    else:
                                        if (key not in self._BVSEparam):
                                            raise Exception("bvse {0}  param can't find!!!!".format(key))
                                            ProgressCall(0)
                                        else:
                                            raise Exception("bvs {0} param can't find!!!!".format(key))
                                            ProgressCall(0)
                        else:
                            ## When Li+, this if-else is for effect between attempt_Li+ and *cations*
                            for (ssymbol, occupyvalue) in site2.GetElementsOccupy().items():
                                if ssymbol != 'Vac':
                                    site2oxi = site2.GetElementsOxiValue()[ssymbol]
                                    
                                    if ssymbol != self._MoveIon:
                                        ## For cations around attempt_Li+, which is not Li+
                                        key = ssymbol + str(site2oxi)
                                        rm2 = self._Elementsparam[key][6]
                                        qm2 = site2oxi / math.sqrt(self._Elementsparam[key][3])
                                        rm1m2 = distance[j][k][dindex]
                                        f = 0.74
                                        if rm1m2 > rm2:
                                            if rm1m2 < Rcutoff:
                                                cdata = cdata + occupyvalue * qm1 * qm2 / rm1m2 * erfc(rm1m2 / (f * (rm1 + rm2))) * qsum
                                                bvelcdata = bvelcdata + occupyvalue * qm1 * qm2 / rm1m2 * (erfc(rm1m2 / (f * (rm1 + rm2))) - erfc(Rcutoff / (f * (rm1 + rm2))))
                                        else:
                                            ## For rm1m2 <= rm2, this means the sites cation is too close to attempt_Li+
                                            ## So the calculation of this attempt_Li+ can be breaked
                                            break_this_point = True
                                            
                                            cdata = 300
                                            bvelcdata = 300
                                    
                                    elif ssymbol == self._MoveIon:
                                        ## For neighbor Li+ around attempt_Li+
                                        key = ssymbol + str(site2oxi)
                                        rm2 = self._Elementsparam[key][6]
                                        qm2 = site2oxi / math.sqrt(self._Elementsparam[key][3])
                                        rm1m2 = distance[j][k][dindex]
                                        f = 0.74
                                        
                                        R_strict = rm2
                                        R_strict = self.Rmin_moveion
                                        # R_strict = 1.45
                                        if rm1m2 > R_strict:
                                            if rm1m2 < Rcutoff:
                                                # cdata = cdata + occupyvalue * qm1 * qm2 / rm1m2 * erfc(rm1m2 / (f * (rm1 + rm2))) * qsum
                                                # bvelcdata = bvelcdata + occupyvalue * qm1 * qm2 / rm1m2 * (erfc(rm1m2 / (f * (rm1 + rm2))) - erfc(Rcutoff / (f * (rm1 + rm2))))
                                                
                                                cdata_Li_i = occupyvalue * qm1 * qm2 / rm1m2 * erfc(rm1m2 / (f * (rm1 + rm2))) * qsum
                                                bvelcdata_Li_i = occupyvalue * qm1 * qm2 / rm1m2 * (erfc(rm1m2 / (f * (rm1 + rm2))) - erfc(Rcutoff / (f * (rm1 + rm2))))
                                                
                                                list_cdata_Li.append([rm1m2, cdata_Li_i])
                                                list_bvelcdata_Li.append([rm1m2, bvelcdata_Li_i])
                                                
                                        else:
                                            # cdata = cdata + 0
                                            # bvelcdata = bvelcdata + 0
                                            
                                            pass
                                    
                    
                    list_cdata_Li.sort(key=lambda x: x[0])
                    list_cdata_Li = list_cdata_Li[1:]
                    list_bvelcdata_Li.sort(key=lambda x: x[0])
                    list_bvelcdata_Li = list_bvelcdata_Li[1:]
                    
                    cdata += sum([i[1] for i in list_cdata_Li])
                    bvelcdata += sum([i[1] for i in list_bvelcdata_Li])
                    
                    self.__Data['BVSE'][i][j][k] = (0.5 * D0 * (data) + 14.4 * cdata)
                    self.__Data['BVS'][i][j][k] = bvsdata
                    self.__Data['BVEL'][i][j][k] = (D0 * (data) + 14.4 * bvelcdata)
        
        self.__Data['BVSE'] = self.__Data['BVSE']  #/self.ValenceOfMoveIon
        self.__Data['BVEL'] = self.__Data['BVEL']  #/self.ValenceOfMoveIon
        for key, data in self.__Data.items():
            self.__Max[key] = np.max(data)
            self.__Min[key] = np.min(data)
            ashape = self.__Data['BVSE'].shape
            self.ndata = np.zeros((2 * ashape[0], 2 * ashape[1], 2 * ashape[2]))
            for i in range(2):
                for j in range(2):
                    for k in range(2):
                        self.ndata[i*ashape[0]:(i+1)*ashape[0], \
                                                 j*ashape[1]:(j+1)*ashape[1], \
                                                 k*ashape[2]:(k+1)*ashape[2]]=\
                                                data-self.__Min[key]
            self.Ea[key] = self.GetEa(self.ndata)

        _BVSE_Ea = [round(i, 2) for i in self.BVSEEa]
        print('\nRes {} | Rmin {} | BVSE Ea {}\n'.format(self._reslen, self.Rmin_moveion, _BVSE_Ea))
        
        # with open('Ea.txt', 'w') as f:
            # f.write('Res {} | Rmin {} | BVSE Ea {}\n'.format(self._reslen, self.Rmin_moveion, _BVSE_Ea))
        
        return True

    def GetEa(self, data):
        return [self.GetConnectEa(data,1),\
                self.GetConnectEa(data,2),\
                self.GetConnectEa(data,3)]

    def GetConnectEa(self, data, connectnumber):
        minvalue = 0.0
        maxvalue = 20.0
        while (maxvalue - minvalue) > 0.01:
            testvalue = (maxvalue + minvalue) / 2.0
            testdata = data < testvalue
            number = self.CaluRegionConnectivity(testdata)
            if number >= connectnumber:
                maxvalue = testvalue
            else:
                minvalue = testvalue
        if maxvalue >= 20.0:
            return 20000
        else:
            return maxvalue

    def CaluRegionConnectivity(self, region):
        struct = scipy.ndimage.generate_binary_structure(3, 3)
        lw, num = measurements.label(region, structure=struct)
        sliced = measurements.find_objects(lw)
        connectnumber = np.zeros(num, dtype=np.int)
        if (num > 0):
            for index, slic in enumerate(sliced):
                for i in range(len(slic)):
                    if ((slic[i].start == 0)
                            and slic[i].stop == region.shape[i]):
                        connectnumber[index] = connectnumber[index] + 1
        return np.max(connectnumber)

    # def write_vti_file(self,filename='test.vti'):
    # imageData = vtk.vtkImageData()
    # imageData.SetDimensions(self._Size)
    # self._reslen=0.1
    # imageData.SetSpacing(self._reslen,self._reslen,self._reslen)
    # imageData.AllocateScalars(vtk.VTK_FLOAT, 1)
    # for k in range(self._Size[2]):
    # for j in range(self._Size[1]):
    # for i in range(self._Size[0]):
    # imageData.SetScalarComponentFromDouble(i, j,k, 0, self.__Data['BVSE'][i][j][k])
    # writer = vtk.vtkXMLImageDataWriter()
    # writer.SetFileName(filename)
    # writer.SetInputData(imageData)
    # writer.Write()

    Data = property(get_data, set_data, del_data, "Data's docstring")
    Max = property(get_max, set_max, del_max, "Max's docstring")
    Min = property(get_min, set_min, del_min, "Min's docstring")


# if __name__=="__main__":
#     bvmdata=loadbvparam('bvmparam.data')
#
#     print(bvmdata['Na'+'1'+'O'+'-2'])



'''
Created on 2017

@author: hebing
'''
import os
import numpy as np
from scipy.special import erfc
from scipy.ndimage import measurements
import scipy.ndimage
import math
from struct import pack


def loadbvparam(filename):
    with open(filename, 'r') as fp:
        lines = fp.readlines()
    bvmpara = {}
    for line in lines:
        varstr = line.split('\t')
        key = varstr[0] + varstr[1] + varstr[2] + varstr[3]
        if key in bvmpara:
            bvmpara[key].append([float(varstr[4]), float(varstr[5])])
        else:
            bvmpara[key] = [[float(varstr[4]), float(varstr[5])]]
    return bvmpara


def loadBVSEparam(filename):
    with open(filename, 'r') as fp:
        lines = fp.readlines()
    BVSEparam = {}
    for line in lines[1:]:
        line = line.strip('\n')
        varstr = line.split('\t')
        key = varstr[0] + varstr[1] + varstr[2] + varstr[3]
        if key in BVSEparam:
            BVSEparam[key].append([
                float(varstr[4]),
                float(varstr[5]),
                float(varstr[6]),
                float(varstr[7]),
                float(varstr[8]),
                float(varstr[9])
            ])
        else:
            BVSEparam[key] = [[
                float(varstr[4]),
                float(varstr[5]),
                float(varstr[6]),
                float(varstr[7]),
                float(varstr[8]),
                float(varstr[9])
            ]]
    return BVSEparam


def loadElementsparam(filename):
    with open(filename, 'r') as fp:
        lines = fp.readlines()
    Elementsparam = {}
    for line in lines:
        line = line.strip('\n')
        varstr = line.split('\t')
        key = varstr[1] + varstr[2]
        Elementsparam[key] = [
            float(varstr[0]),
            float(varstr[3]),
            float(varstr[4]),
            float(varstr[5]),
            float(varstr[6]),
            float(varstr[7]),
            float(varstr[8]),
            float(varstr[9])
        ]
    return Elementsparam


class BVAnalysis(object):
    '''
    for bond valence method analyze ion channel
    '''

    def __init__(self, struc=None, size=[50, 50, 50]):
        '''
        Constructor
        '''
        self._Struct = struc
        self._Size = [50, 50, 50]
        self.__Data = {}
        self.__Max = {}
        self.__Min = {}
        self.Ea = {}
        self._MoveIon = 'Li'
        self._poss = None
        self.ValenceOfMoveIon = 1
        self.stop = False
        module_dir = os.path.dirname(os.path.abspath(__file__))

        self._BVparam = loadbvparam(os.path.join(module_dir, 'bvmparam.dat'))
        self._BVSEparam = loadBVSEparam(os.path.join(module_dir, 'bvse.dat'))
        self._Elementsparam = loadElementsparam(
            os.path.join(module_dir, 'elements.dat'))

    def get_data(self):
        return self.__Data

    def get_max(self):
        return self.__Max

    def get_min(self):
        return self.__Min

    def set_data(self, value):
        self.__Data = value

    def set_max(self, value):
        self.__Max = value

    def set_min(self, value):
        self.__Min = value

    def del_data(self):
        del self.__Data

    def del_max(self):
        del self.__Max

    def del_min(self):
        del self.__Min

    @property
    def BVSMin(self):
        return self.__Min['BVS']

    @property
    def BVSEMin(self):
        return self.__Min['BVSE']

    @property
    def BVELMin(self):
        return self.__Min['BVEL']

    @property
    def BVSMax(self):
        return self.__Max['BVS']

    @property
    def BVSEMax(self):
        return self.__Max['BVSE']

    @property
    def BVELMax(self):
        return self.__Max['BVEL']

    @property
    def BVSEEa(self):
        return self.Ea['BVSE']

    @property
    def BVELEa(self):
        return self.Ea['BVEL']

    def SetStructure(self, struc):
        self._Struct = struc

    def SetMoveIon(self, moveion):
        self._MoveIon = moveion

    def SetResolution(self, size):
        self._Size = size

    def SetLengthResolution(self, reslen=0.1):
        abc = self._Struct.Getabc()
        self._reslen = reslen
        self._Size[0] = int(abc[0] / reslen) + 1
        self._Size[1] = int(abc[1] / reslen) + 1
        self._Size[2] = int(abc[2] / reslen) + 1

    def SaveBVSEData(self, filename):
        self.SaveBinaryData(filename + '_BVSE.pgrid', self.__Data['BVSE'],
                            'BVSE')
        np.save(filename + "_BVSE.npy", self.__Data['BVSE'])

    def SaveBVELData(self, filename):
        self.SaveBinaryData(filename + '_BVEL.pgrid', self.__Data['BVEL'],
                            'BVEL')

    def SaveBVSData(self, filename):
        self.SaveBinaryData(filename + '_BVS.pgrid', self.__Data['BVS'], 'BVS')

    def SaveData(self, filename, data, datatype=' '):
        fp = open(filename, 'w')
        fp.write(datatype + '\n')
        alpha = self._Struct.GetAlphaBetaGama()
        abc = self._Struct.Getabc()
        fp.write(str(abc[0]) + ' ' + str(abc[1]) + ' ' + str(abc[2]) + ' ')
        fp.write(
            str(alpha[0]) + ' ' + str(alpha[1]) + ' ' + str(alpha[2]) + '\n')
        fp.write(
            str(self._Size[0]) + ' ' + str(self._Size[1]) + ' ' +
            str(self._Size[2]) + '\n ')
        for i in range(self._Size[0]):
            for j in range(self._Size[1]):
                for k in range(self._Size[2]):
                    fp.write(str(data[i][j][k]) + '\n')
        fp.close()

    def SaveBinaryData(self, filename, data, datatype=' '):
        latticeparam = self._Struct.Getabc() + self._Struct.GetAlphaBetaGama()
        with open(filename, 'wb') as outf:
            outf.write(pack('4i', 3, 0, 0, 0))  # file format version
            outf.write(pack(
                '80s',
                datatype.ljust(80).encode('utf-8')))  # Title 80 characters
            outf.write(pack('i', 1))  # gType 0 for *.ggrid, 1 for *.pgrid
            outf.write(pack('i', 0))  # fType 0
            outf.write(pack('i', 1))  # nVal dummy
            outf.write(pack('i', 3))  # dimension
            outf.write(pack('3i', *self._Size))  # numbers of voxels
            outf.write(pack('i', data.size))  # Total number of voxels
            outf.write(pack('6f', *latticeparam))  # lattice parameters
            for k in range(self._Size[2]):
                for j in range(self._Size[1]):
                    for i in range(self._Size[0]):
                        outf.write(pack('f', data[i][j][k]))

    def ReadData(self, filename, data, datatype=' '):
        fp = open(filename, 'r')
        datatype = fp.readline()
        alpha = self._Struct.GetAlphaBetaGama()
        abc = self._Struct.Getabc()
        abc = fp.readline().split()
        alpha = fp.readline().split()
        strs = fp.readline().split()
        size = [int(i) for i in strs]
        if datatype == 'bvs':
            self.__Data['BVS'] = np.zeros(size)
            data = self.__Data['BVS']
        else:
            self.__Data['BVSE'] = np.zeros(size)
            data = self.__Data['BVSE']
        for i in range(size[2]):
            for j in range(size[1]):
                for k in range(size[0]):
                    strs = fp.readline().split()
                    data[k][j][i] = int(strs[0])

        fp.close()

    def GetBVSData(self):
        return self.__Data['BVS']

    def GetBVSEData(self):
        return self.__Data['BVSE']

    def GetPosSet(self):
        return self._poss

    def CaluGlobalIndex(self, ):
        self.stop = False
        self._Size.reverse()
        centrepos = np.zeros(self._Size + [3])
        self._Data = np.zeros(self._Size, dtype=np.double)
        for k in range(self._Size[0]):
            for j in range(self._Size[1]):
                for i in range(self._Size[2]):
                    centrepos[k][j][i][2] = k / (self._Size[0] - 1.0)
                    centrepos[k][j][i][1] = j / (self._Size[1] - 1.0)
                    centrepos[k][j][i][0] = i / (self._Size[2] - 1.0)

        self._poss = self._Struct.FracPosToCartPos(centrepos)
        (distance, neighborsindex) = self._Struct.GetKNeighbors(self._poss,
                                                                kn=8)

        site2atoms = None
        for k in range(self._Size[0]):

            for j in range(self._Size[1]):
                for i in range(self._Size[2]):
                    if self.stop:
                        return
                    for dindex, index in enumerate(neighborsindex[k][j][i]):
                        site2 = self._Struct.GetSuperCellusites()[index]
                        if site2.GetIronType() * self.ValenceOfMoveIon < 0:
                            for key in site2.GetElementsOccupy():
                                if key != 'Vac':
                                    site2atoms = key
                            if self.ValenceOfMoveIon > 0:
                                key = "".join([
                                    self._MoveIon,
                                    str(self.ValenceOfMoveIon), site2atoms,
                                    str(site2.GetElementsOxiValue()
                                        [site2atoms])
                                ])
                            else:
                                key = "".join([
                                    site2atoms,
                                    str(site2.GetElementsOxiValue()
                                        [site2atoms]), self._MoveIon,
                                    str(self.ValenceOfMoveIon)
                                ])
                            if key in self._BVparam:
                                (r, b) = self._BVparam[key][0]
                                bv = np.exp(
                                    (r - distance[k][j][i][dindex]) / b)
                                self.__Data['BVS'][k][j][
                                    i] = self.__Data['BVS'][k][j][i] + bv
                            else:
                                print('Not Find bvs param' + key)
                        else:
                            pass
                            #print(site2.GetSiteLabel())
    def CaluBVSE(self, ProgressCall):
        if len(self._Struct._OxidationTable) == 0:
            raise Exception(
                "can't caculation bvs and Ea without atom oxi info !")
        self.stop = False
        centrepos = np.zeros(self._Size + [3])
        self.__Data['BVS'] = np.zeros(self._Size, dtype=np.double)
        self.__Data['BVSE'] = np.zeros(self._Size, dtype=np.double)
        self.__Data['BVEL'] = np.zeros(self._Size, dtype=np.double)
        for k in range(self._Size[2]):
            for j in range(self._Size[1]):
                for i in range(self._Size[0]):
                    centrepos[i][j][k][2] = k / (self._Size[2] - 1.0)
                    centrepos[i][j][k][1] = j / (self._Size[1] - 1.0)
                    centrepos[i][j][k][0] = i / (self._Size[0] - 1.0)
        if ProgressCall:
            ProgressCall(10)
        self._poss = self._Struct.FracPosToCartPos(centrepos)

        atomsq = {}
        for atom in self._Struct._atomsymbols:
            atomsq[atom] = 0
            for site in self._Struct._Sites:
                if atom in site.GetElements():
                    key = atom + str(site.GetElementsOxiValue()[atom])
                    atomsq[atom] = atomsq[atom] + site.GetElementsOxiValue(
                    )[atom] * site.GetElementsOccupy()[atom] / math.sqrt(
                        self._Elementsparam[key][3])
        qsumanion = 0
        qsumcation = 0
        for (atom, value) in atomsq.items():
            if value > 0:
                qsumcation = qsumcation + value
            elif value < 0:
                qsumanion = qsumanion + value
            else:
                qsumanion = 0
                qsumcation = 0
        qsum = 0.0
        if self.ValenceOfMoveIon > 0:
            qsum = -qsumanion / qsumcation
        else:
            qsum = -qsumcation / qsumanion
        key1 = self._MoveIon + str(self.ValenceOfMoveIon)
        qm1 = self.ValenceOfMoveIon / math.sqrt(self._Elementsparam[key1][3])
        rm1 = self._Elementsparam[key1][6]
        for i in range(self._Size[0]):
            (distance,
             neighborsindex) = self._Struct.GetKNeighbors(self._poss[i],
                                                          kn=128)
            if ProgressCall:
                ProgressCall(10 + i * 90 / (self._Size[0] - 1))
            for j in range(self._Size[1]):
                for k in range(self._Size[2]):
                    if self.stop:
                        return False
                    bvsdata = 0.0
                    data = 0.0
                    cdata = 0.0
                    bvelcdata = 0.0
                    N = 0
                    D0 = 0.0
                    Rcutoff = 10.0
                    alpha = 0.0
                    occupyvalue = 0.0
                    for dindex, index in enumerate(neighborsindex[j][k]):
                        site2 = self._Struct._SuperCellusites[index]
                        if site2.GetIronType() * self.ValenceOfMoveIon < 0:
                            for (ssymbol, occupyvalue
                                 ) in site2.GetElementsOccupy().items():
                                if ssymbol != 'Vac':
                                    site2oxi = site2.GetElementsOxiValue(
                                    )[ssymbol]
                                    if self.ValenceOfMoveIon > 0:
                                        key = "".join([
                                            self._MoveIon,
                                            str(self.ValenceOfMoveIon),
                                            ssymbol,
                                            str(site2oxi)
                                        ])
                                    else:
                                        key = "".join([
                                            ssymbol,
                                            str(site2oxi), self._MoveIon,
                                            str(self.ValenceOfMoveIon)
                                        ])
                                    if (key in self._BVSEparam) and (
                                            key in self._BVparam):
                                        (Nc, r0, Rcutoff, D0, Rmin,
                                         alpha) = self._BVSEparam[key][0]
                                        (r, b) = self._BVparam[key][0]
                                        Rcutoff = 10.0
                                        if distance[j][k][dindex] <= Rcutoff:
                                            # bv=np.exp((r0-distance[k][j][i][dindex])*alpha)
                                            bvs = np.exp(
                                                (r - distance[j][k][dindex]) /
                                                b)
                                            smin = np.exp(
                                                (Rmin - distance[j][k][dindex])
                                                * alpha)
                                            en_s = np.exp(
                                                (Rmin - Rcutoff) * alpha)
                                            bvsdata = bvsdata + bvs  #*occupyvalue-((en_s-1)**2-1)
                                            data = data + ((smin - 1)**2 -
                                                           1) * occupyvalue
                                    else:
                                        if (key not in self._BVSEparam):
                                            raise Exception(
                                                "bvse {0}  param can't find!!!!"
                                                .format(key))
                                            ProgressCall(0)
                                        else:
                                            raise Exception(
                                                "bvs {0} param can't find!!!!".
                                                format(key))
                                            ProgressCall(0)
                        else:
                            for (ssymbol, occupyvalue
                                 ) in site2.GetElementsOccupy().items():
                                if ssymbol != 'Vac':
                                    site2oxi = site2.GetElementsOxiValue(
                                    )[ssymbol]
                                    if ssymbol != self._MoveIon:
                                        key = ssymbol + str(site2oxi)
                                        rm2 = self._Elementsparam[key][6]
                                        qm2 = site2oxi / math.sqrt(
                                            self._Elementsparam[key][3])
                                        rm1m2 = distance[j][k][dindex]
                                        f = 0.74
                                        if rm1m2 > rm2:
                                            if rm1m2 < Rcutoff:
                                                cdata = cdata + occupyvalue * qm1 * qm2 / rm1m2 * erfc(
                                                    rm1m2 /
                                                    (f * (rm1 + rm2))) * qsum
                                                bvelcdata = bvelcdata + occupyvalue * qm1 * qm2 / rm1m2 * (
                                                    erfc(rm1m2 /
                                                         (f * (rm1 + rm2))) -
                                                    erfc(Rcutoff /
                                                         (f * (rm1 + rm2))))
                                        else:
                                            cdata = 300
                                            bvelcdata = 300
                    self.__Data['BVSE'][i][j][k] = (0.5 * D0 * (data) +
                                                    14.4 * cdata)
                    self.__Data['BVS'][i][j][k] = bvsdata
                    self.__Data['BVEL'][i][j][k] = (D0 * (data) +
                                                    14.4 * bvelcdata)
        self.__Data['BVSE'] = self.__Data['BVSE']  #/self.ValenceOfMoveIon
        self.__Data['BVEL'] = self.__Data['BVEL']  #/self.ValenceOfMoveIon
        for key, data in self.__Data.items():
            self.__Max[key] = np.max(data)
            self.__Min[key] = np.min(data)
            ashape = self.__Data['BVSE'].shape
            self.ndata = np.zeros(
                (2 * ashape[0], 2 * ashape[1], 2 * ashape[2]))
            for i in range(2):
                for j in range(2):
                    for k in range(2):
                        self.ndata[i*ashape[0]:(i+1)*ashape[0], \
                                                 j*ashape[1]:(j+1)*ashape[1], \
                                                 k*ashape[2]:(k+1)*ashape[2]]=\
                                                data-self.__Min[key]
            self.Ea[key] = self.GetEa(self.ndata)

        #print('BVSE Ea is {0},\nBVEL Ea is {1}'.format(self.BVSEEa,self.BVELEa))
        return True

    def GetEa(self, data):
        return [self.GetConnectEa(data,1),\
                self.GetConnectEa(data,2),\
                self.GetConnectEa(data,3)]

    def GetConnectEa(self, data, connectnumber):
        minvalue = 0.0
        maxvalue = 20.0
        while (maxvalue - minvalue) > 0.01:
            testvalue = (maxvalue + minvalue) / 2.0
            testdata = data < testvalue
            number = self.CaluRegionConnectivity(testdata)
            if number >= connectnumber:
                maxvalue = testvalue
            else:
                minvalue = testvalue
        if maxvalue >= 20.0:
            return 20000
        else:
            return maxvalue

    def CaluRegionConnectivity(self, region):
        struct = scipy.ndimage.generate_binary_structure(3, 3)
        lw, num = measurements.label(region, structure=struct)
        sliced = measurements.find_objects(lw)
        connectnumber = np.zeros(num, dtype=np.int)
        if (num > 0):
            for index, slic in enumerate(sliced):
                for i in range(len(slic)):
                    if ((slic[i].start == 0)
                            and slic[i].stop == region.shape[i]):
                        connectnumber[index] = connectnumber[index] + 1
        return np.max(connectnumber)

    # def write_vti_file(self,filename='test.vti'):
    # imageData = vtk.vtkImageData()
    # imageData.SetDimensions(self._Size)
    # self._reslen=0.1
    # imageData.SetSpacing(self._reslen,self._reslen,self._reslen)
    # imageData.AllocateScalars(vtk.VTK_FLOAT, 1)
    # for k in range(self._Size[2]):
    # for j in range(self._Size[1]):
    # for i in range(self._Size[0]):
    # imageData.SetScalarComponentFromDouble(i, j,k, 0, self.__Data['BVSE'][i][j][k])
    # writer = vtk.vtkXMLImageDataWriter()
    # writer.SetFileName(filename)
    # writer.SetInputData(imageData)
    # writer.Write()

    Data = property(get_data, set_data, del_data, "Data's docstring")
    Max = property(get_max, set_max, del_max, "Max's docstring")
    Min = property(get_min, set_min, del_min, "Min's docstring")


# if __name__=="__main__":
#     bvmdata=loadbvparam('bvmparam.data')
#
#     print(bvmdata['Na'+'1'+'O'+'-2'])



# Copyright (c) PyBW
# Distributed under the terms of the MIT License.

import sys
import platform
from setuptools import setup, find_packages

is_win_64 = sys.platform.startswith('win') and platform.machine().endswith('64')
extra_link_args = ['-Wl,--allow-multiple-definition'] if is_win_64 else []

with open('README.md', encoding='utf-8') as f:
    readme = f.read()


setup(
    name='pybw',
    version='24.6.15.1',
    python_requires='>=3.6',
    
    author='Bowei Pu',
    author_email='pubowei@foxmail.com',
    
    description='pybw',
    long_description=readme, 
    long_description_content_type='text/markdown',
    url='https://gitee.com/pubowei/pybw',
    keywords=['pybw', 'tools'],
    license='MIT',
    
    packages=find_packages(), 
    include_package_data=True, 
    
    entry_points={
        'console_scripts': ['ccnbmod = pybw.ccnb_mod.ccnb_mod:main']
    }, 

    project_urls={
        'Docs': 'https://gitee.com/pubowei/pybw',
        'Package': 'https://pypi.org/project/pybw',
        'Repo': 'https://gitee.com/pubowei/pybw',
    },


    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.6',
    ],
)


"""
setup(
    name='pybw',
    version='2023.2.28',
    python_requires='>=3.6',
    
    author='Bowei Pu',
    author_email='pubowei@foxmail.com',
    maintainer='Bowei Pu',
    maintainer_email='pubowei@foxmail.com',
    
    description='pybw',
    long_description=readme, 
    long_description_content_type='text/markdown',
    url='https://pubowei.cn',
    keywords=['pybw', 'tools'],
    license='MIT',
    
    packages = find_packages(), 
    
    entry_points={
        'console_scripts': ['ccnbmod = pybw:ccnbmod.ccnbmod.main']
    }, 
    scripts=['pybw/ccnbmod/ccnbmod.py'], 
    
    
    install_requires=[
        'pathlib>=1.0.1',
        'tqdm',
    ],
    extras_require={
        'ase': ['ase>=1.0'],
    },
    project_urls={
        'Docs': 'https://pubowei.cn',
        'Package': 'https://pypi.org/project/pybw',
        'Repo': 'https://gitee.com/pubowei/pybw',
    },

    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering :: Chemistry',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Scientific/Engineering :: Physics',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)
"""



# -*- coding: utf-8 -*-
'''
modify author: Bowei Pu
version: 2022.08.17
note: This is a modified module from BVSE program by Bing He
function: __init__ class of bvseLi
'''
import os
import time
from pathlib import Path
import shutil
from pickle import NONE
import ase.io
from rich.progress import Progress

import bvseLi.Structure as Structure
import bvseLi.BVAnalysis as BVAnalysis
import bvseLi.BVAnalysisLi as BVAnalysisLi


class Progressbar:

    def __init__(self):
        self.progress = Progress()
        self.task = None

    def set_task(self, workstr: str, total):
        self.task = self.progress.add_task(workstr, total)

    def updatetask(self, i: int):
        self.progress.update(self.task, advance=0.9)
        print('    bvse computing  {0:.2f}%'.format(i))


def time_used(time0, if_print=False, return_type='tuple'):
    dt = int(time.time() - time0)
    h, m, s = dt // 3600, dt // 60 % 60, dt % 60
    if if_print:
        print('    time used: {:0>2}h {:0>2}m {:0>2}s'.format(h, m, s))
        # print('    time used: {:0>2}:{:0>2}:{:0>2}'.format(h, m, s))
    if return_type.lower() in ['default', 'normal', 'tuple', 'list']:
        return h, m, s
    elif return_type.lower() in ['dic']:
        return {'h': h, 'm': m, 's': s}
    else:
        return


def time_used_old(time0, if_print=False):
    dt = time.time() - time0
    sec = int(dt)
    mins = round(sec / 60, 2)
    hour = round(sec / 60 / 60, 2)
    if if_print:
        print('    time used: {} h, {} m, {} s'.format(hour, mins, sec))
    dic = {'h': hour, 'm': mins, 's': sec}
    return dic


class OsPath():
    '''
    '''
    def __init__(self, file):
        self.file = file
        
        self.path = self.abspath = os.path.abspath(self.file)
        self.parent, self.name = os.path.split(self.abspath)
        self.dirname, self.basename, self.filename = self.parent, self.name, self.name
        self.stem, self.suffix = os.path.splitext(self.name)
        self.path_stem = self.abs_stem = os.path.splitext(self.abspath)[0]
        
        self.str = self.string = self.abspath
        self.posix = self.as_posix()
    
    def __str__(self):
        return self.abspath
    
    def __repr__(self):
        return self.__str__()
    
    def __fspath__(self):
        return self.abspath
    
    def absolute(self):
        return self.abspath
    
    def as_posix(self):
        return Path(self.abspath).as_posix()
    
    @staticmethod
    def get_posix(cls, path):
        return Path(path).as_posix()
    
    def join(self, other: str):
        return self.__class__(os.path.join(self.abspath, other))
    
    def joinpath(self, other: str):
        return self.join(other)
    
    @property
    def exists(self):
        return os.path.exists(self.abspath)
        
    @property
    def isfile(self):
        return os.path.isfile(self.abspath)
    
    @property
    def isdir(self):
        return os.path.isdir(self.abspath)
    
    def rename(self, new_name):
        return self.__class__(os.path.join(self.dirname, new_name))


def bv_calculation(filename,
                   moveion='Li',
                   valenceofmoveion=1,
                   resolution=0.1,
                   progress=None, 
                   Rmin_moveion=3, 
                   engine='bvseLi',
                   pack_files=True, 
                   skip_done=True, 
                   rename_cif=True
                  ):
    '''
    Convenient function to apply BVSELi calculation
    
    Args:
        filename: file path
        moveion: mobile ion
        valenceofmoveion: valence of mobile ion
        resolution: resolution
        progress: whether desplay progress bar when cal
        Rmin_moveion: Rmin cutoff to judge whether cal nearby mobile ion
        engine: choose different bv method. 'bvse' or 'bvseLi'
        pack_files: if pack output files after calculation
        skip_done: if skip .cif file which is startswith 'done--'
        rename_cif: if rename '*.cif' file to 'done--*.cif' after calculation
    
    Returns:
        bvs.Ea['BVSE']
    
    Outputs:
        .npy: Data for numpy load
        .pgrid: Grid data for VESTA
        .grd: Grd data
        INFO.csv: Log to record related info, include filename, res, Rmin, Ea and etc.
    '''
    if skip_done:
        if Path(filename).name.startswith('done--'):
            return
    
    if not progress or progress.lower() in ['default', 'auto']:
        prgbar = Progressbar()
        prgbar.set_task('bvse computing progress...', total=100)
        progress = prgbar.updatetask
    
    if engine.lower() in ['default', 'normal', 'bvse']:
        engine = 'bvse'
    elif engine.lower() in ['bvseli']:
        engine = 'bvseli'
    
    atoms = ase.io.read(filename, store_tags=True)
    struc = Structure.Structure()
    struc.GetAseStructure(atoms)
    
    if engine == 'bvse':
        bvs = BVAnalysis.BVAnalysis()
        BVMethod = 'BVSE'
    elif engine == 'bvseli':
        bvs = BVAnalysisLi.BVAnalysis()
        BVMethod = 'BVSELi'
    
    bvs.SetStructure(struc)
    bvs.SetMoveIon(moveion)
    bvs.ValenceOfMoveIon = valenceofmoveion
    bvs.SetLengthResolution(resolution)
    
    time0 = time.time()
    progress = None
    if engine == 'bvse':
        bvs.CaluBVSE(progress)
    elif engine == 'bvseli':
        bvs.CaluBVSE(progress, Rmin_moveion)
    time_info = time_used(time0, if_print=False, return_type='dic')
    
    file = OsPath(filename)
    file_name = Path(filename).name
    filepath = os.path.abspath(filename)
    filepath_dire = os.path.split(filepath)[0]
    filepath_stem = os.path.splitext(filepath)[0]
    
    bvs.SaveBVSEData(file.abs_stem)
    bvs.SaveBVELData(file.abs_stem)
    bvs.SaveBVSData(file.abs_stem)
    
    bv_data = bvs.get_data()
    bvs.SaveData('{}_BVSELi.grd'.format(file.abs_stem), bv_data['BVSE'], 'BVSE')
    
    with open('{}_INFO.csv'.format(file.abs_stem), 'w') as f:
        f.write('filename,resolution,Rmin_moveion,Ea_BVSE,time_used\n')
        _BVSE_Ea = [round(i, 2) for i in bvs.BVSEEa]
        _BVSE_Ea = '[{}]'.format(' '.join([str(i) for i in _BVSE_Ea]))
        _time_info = '{}h {}m {}s'.format(time_info['h'], time_info['m'], time_info['s'])
        f.write('{0},{1},{2},{3},{4}'.format(filename, bvs._reslen, bvs.Rmin_moveion, _BVSE_Ea, _time_info))
    
    if pack_files:
        date_short = time.strftime('%y%m%d_%H%M%S')
        _dire_name = '{}_BVSELi--Res_{}-Rmin_{}--{}'.format(date_short, bvs._reslen, bvs.Rmin_moveion, file.name)
        _dire = os.path.join(file.dirname, _dire_name)
        os.mkdir(_dire)
        shutil.copy2(file.abspath, _dire)
        for _suffix in ['_BVSELi.npy', 
                        '_BVSELi.pgrid', 
                        '_BVELLi.pgrid', 
                        '_BVSLi.pgrid', 
                        '_BVSELi.grd', 
                        '_INFO.csv'
                       ]:
            shutil.move('{}{}'.format(file.abs_stem, _suffix), _dire)
    
    if rename_cif:
        file_new = file.rename('done--' + file.name).string
        os.rename(filename, file_new)
    
    return bvs.Ea['BVSE']


def bv_calculation_normal(filename,
                          moveion='Li',
                          valenceofmoveion=1,
                          resolution=0.1,
                          progress='default', 
                          Rmin_moveion=3
                         ):
    
    prgbar = Progressbar()
    prgbar.set_task('bvse computing progress...', total=100)
    if progress in ['default', 'auto']:
        progress = prgbar.updatetask
    
    atoms = ase.io.read(filename, store_tags=True)
    struc = Structure.Structure()
    struc.GetAseStructure(atoms)
    
    bvs = BVAnalysis.BVAnalysis()
    bvs.SetStructure(struc)
    bvs.SetMoveIon(moveion)
    bvs.ValenceOfMoveIon = valenceofmoveion
    bvs.SetLengthResolution(resolution)
    
    # bvs.CaluBVSE(progress, Rmin_moveion)
    bvs.CaluBVSE(progress)
    
    bvs.SaveBVSEData(os.path.splitext(filename)[0])
    bv_data = bvs.get_data()
    bvs.SaveData(os.path.splitext(filename)[0] + '.bvse.grd', bv_data['BVSE'], 'BVSE')
    
    return bvs.Ea['BVSE']


def bv_calculation_old(filename,
                   moveion='Li',
                   valenceofmoveion=1,
                   resolution=0.1,
                   progress=None):
    atoms = ase.io.read(filename, store_tags=True)
    struc = Structure.Structure()
    struc.GetAseStructure(atoms)
    bvs = BVAnalysis.BVAnalysis()
    bvs.SetStructure(struc)
    bvs.SetMoveIon(moveion)
    bvs.ValenceOfMoveIon = valenceofmoveion
    bvs.SetLengthResolution(resolution)
    
    bvs.CaluBVSE(progress)
    
    bvs.SaveBVSEData(os.path.splitext(filename)[0])
    bv_data = bvs.get_data()
    bvs.SaveData(os.path.splitext(filename)[0] + '.bvse.grd', bv_data['BVSE'], 'BVSE')
    
    return bvs.Ea['BVSE']
    
    
    
    
    